from pathlib import Path

from simple_scenario import Vehicle
from simple_scenario.rendering import create_scenario_plot_ax
from simple_scenario.road import Road, StraightSegment


class TestVehicle:
    RESULT_DIR = Path(__file__).parent / "results" / "test_vehicle"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_vehicle(self):
        result_dir = self.RESULT_DIR / "test_vehicle"
        result_dir.mkdir(exist_ok=True)

        print("Create vehicle")

        vehicle = Vehicle(0, 1000, 10, 0, 27.78)

        # Access parameters
        print(f"Length: {vehicle.length}")
        print(f"Width: {vehicle.width}")

        # For compilation, we need a road
        print("Compile vehicle")
        road = Road(3, 3.75, [StraightSegment(300)])

        duration = 10
        dt = 0.1
        vehicle.compile(road, duration, dt)

        vehicle.render(result_dir, plot_name="vehicle_default")
        vehicle.render(result_dir, plot_name="vehicle_clean", clean=True)

        with create_scenario_plot_ax(result_dir, "vehicle_on_road_default") as ax:
            vehicle.render(ax)
            road.render(ax)
        with create_scenario_plot_ax(
            result_dir, "vehicle_on_road_clean", clean=True
        ) as ax:
            vehicle.render(ax)
            road.render(ax)

    def test_vehicle_lc_left(self):
        result_dir = self.RESULT_DIR / "test_vehicle_lc_left"
        result_dir.mkdir(exist_ok=True)

        print("Create vehicle")

        vehicle = Vehicle(0, 1000, 10, 0, 27.78, lc_direction=1, lc_delay=2)

        # Access parameters
        print(f"Length: {vehicle.length}")
        print(f"Width: {vehicle.width}")

        # For compilation, we need a road
        print("Compile vehicle")
        road = Road(3, 3.75, [StraightSegment(300)])

        duration = 10
        dt = 0.1
        vehicle.compile(road, duration, dt)

        vehicle.render(result_dir, plot_name="vehicle_lc_left_default")
        vehicle.render(result_dir, plot_name="vehicle_lc_left_clean", clean=True)

        with create_scenario_plot_ax(
            result_dir, "vehicle_lc_left_on_road_default"
        ) as ax:
            vehicle.render(ax)
            road.render(ax)
        with create_scenario_plot_ax(
            result_dir, "vehicle_lc_left_on_road_clean", clean=True
        ) as ax:
            vehicle.render(ax)
            road.render(ax)

    def test_vehicle_lc_right(self):
        result_dir = self.RESULT_DIR / "test_vehicle_lc_right"
        result_dir.mkdir(exist_ok=True)

        print("Create vehicle")

        vehicle = Vehicle(0, 1001, 10, 0, 27.78, lc_direction=-1, lc_delay=2)

        # Access parameters
        print(f"Length: {vehicle.length}")
        print(f"Width: {vehicle.width}")

        # For compilation, we need a road
        print("Compile vehicle")
        road = Road(3, 3.75, [StraightSegment(300)])

        duration = 10
        dt = 0.1
        vehicle.compile(road, duration, dt)

        vehicle.render(result_dir, plot_name="vehicle_lc_right_default")
        vehicle.render(result_dir, plot_name="vehicle_lc_right_clean", clean=True)

        with create_scenario_plot_ax(
            result_dir, "vehicle_lc_right_on_road_default"
        ) as ax:
            vehicle.render(ax)
            road.render(ax)
        with create_scenario_plot_ax(
            result_dir, "vehicle_lc_right_on_road_clean", clean=True
        ) as ax:
            vehicle.render(ax)
            road.render(ax)


if __name__ == "__main__":
    tester = TestVehicle()
    tester.test_vehicle()
    tester.test_vehicle_lc_left()
    tester.test_vehicle_lc_right()



import numpy as np

from loguru import logger

from .road_segment import RoadSegment


class StraightSegment(RoadSegment):
    def __init__(self, length: float, heading: float = 0.0) -> None:
        """
        length: Length of the road segment in m
        heading: Heading of the road segment in rad
        """

        self._config = {"length": length, "heading": heading}

        super().__init__()

        self._length = length
        self._heading = heading

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} length={self._length}>"

    @property
    def config(self) -> dict:
        return self._config

    @property
    def length(self) -> float:
        return self._length

    @property
    def heading(self) -> float:
        return self._heading

    def compute_ref_line(self, start_x: float, start_y: float) -> np.ndarray:
        logger.debug(
            "Create ref_line of straight segment starting at: {}, {}", start_x, start_y
        )

        straight_x = [start_x]
        straight_y = [start_y]

        for _ in np.arange(
            self._sampling_distance,
            self._length + self._sampling_distance,
            self._sampling_distance,
        ):
            next_straight_x = straight_x[-1] + self._sampling_distance * np.cos(
                self._heading
            )
            next_straight_y = straight_y[-1] + self._sampling_distance * np.sin(
                self._heading
            )

            straight_x.append(next_straight_x)
            straight_y.append(next_straight_y)

        ref_line = np.array([straight_x, straight_y]).transpose()

        self._ref_line = ref_line
        self._ref_line_curvature = np.zeros_like(ref_line[:, 0])
        self._ref_line_heading = self._heading * np.ones_like(ref_line[:, 0])

        return ref_line

    def compute_offset_line(self, lateral_offset: float) -> np.ndarray:
        x = []
        y = []

        for p in self._ref_line:
            x.append(p[0] + lateral_offset * np.cos(self._heading - np.pi / 2))
            y.append(p[1] + lateral_offset * np.sin(self._heading - np.pi / 2))

        offset_line = np.array([x, y]).transpose()

        self._offset_lines[lateral_offset] = offset_line

        return offset_line



from __future__ import annotations

import sys
import typer

from loguru import logger
from pathlib import Path

from .scenario import Scenario

app = typer.Typer(pretty_exceptions_show_locals=False)


@app.command()
def hello_world() -> None:
    print("hello world!")


@app.command()
def create(config_file: str, result_dir: str = None, osc: bool = True) -> None:  # noqa: RUF013
    logger.configure(handlers=[{"sink": sys.stdout, "level": "WARNING"}])

    config_file = Path(config_file)

    if not config_file.is_file():
        msg = "Please provide a valid config file."
        raise FileNotFoundError(msg)

    if result_dir is None:
        result_dir = config_file.parent
    result_dir = Path(result_dir)

    scenario = Scenario.from_config_file(config_file)

    scenario_result_dir = result_dir / f"{scenario.id}_openx"
    scenario_result_dir.mkdir(exist_ok=True)

    if osc:
        print(f"Create openX scenario at: {scenario_result_dir.resolve()}/")

        scenario.save(scenario_result_dir, mode="openx")

        xosc_file = scenario_result_dir / f"{scenario.id}.xosc"

        print(f"To run use: esmini --window --osc {xosc_file.resolve()}")


if __name__ == "__main__":
    app()



from __future__ import annotations

import numpy as np

from copy import deepcopy
from functools import lru_cache
from typing import TYPE_CHECKING

from commonroad_dc.geometry.util import resample_polyline
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem

if TYPE_CHECKING:
    from commonroad.scenario.lanelet import Lanelet, LaneletNetwork

from .neighbour_lanes import NeighbourLanes
from .surrounding_vehicles import SurroundingVehicles
from .traffic_sign_interpreter_fixed import TrafficSigInterpreterFixed
from .out_of_projection_domain_exception import OutOfProjectionDomainError


class LaneletNetworkWrapper:
    """
    Wrapper around CR laneletnetwork with easier handling of frenet coordinate systems.
    """

    def __init__(self, lanelet_network: LaneletNetwork) -> None:
        self._lanelet_network = lanelet_network

        self._lanelet_list = self._lanelet_network.lanelets
        # Assumption: leftmost llt is the last in the list.
        self._ref_line = self._lanelet_list[-1].left_vertices

        self._lanelets_by_llt_id = {
            llt.lanelet_id: llt for llt in self._lanelet_network.lanelets
        }

        self._frenet_frames = {"ref": self._create_frenet_frame(self._ref_line)}

        self._traffic_sign_interpreter = TrafficSigInterpreterFixed(
            "ZAM", self._lanelet_network
        )

    # -- General access to the map --

    @property
    def lanelet_list(self) -> list:
        return self._lanelet_list

    def get_lanelet_dict(self) -> dict:
        lanelet_dict = {llt.lanelet_id: llt for llt in self._lanelet_list}
        return lanelet_dict

    # -- Map matching ---

    @lru_cache(maxsize=100000)
    def find_lanelet_id_by_position(self, x: float, y: float) -> int:
        # Assumption: there are not overlapping lanelets
        llt_id = self._lanelet_network.find_lanelet_by_position([[x, y]])[0][0]

        return llt_id

    def get_lanelet_by_id(self, llt_id: int) -> Lanelet:
        self._check_llt_id(llt_id)

        return self._lanelets_by_llt_id[llt_id]

    def get_speed_limit_by_pos(self, x: float, y: float) -> float:
        lanelet_id = self.find_lanelet_id_by_position(x, y)
        speed_limit = self._traffic_sign_interpreter.speed_limit(
            frozenset([lanelet_id])
        )

        return speed_limit

    # -- Neighbour lanes --

    def find_available_neighbour_lanes(self, ego_lanelet_id: int) -> NeighbourLanes:
        lanelet = self.get_lanelet_by_id(ego_lanelet_id)

        left_lane_available = False
        left_lane_lanelet_id = -1
        if lanelet.adj_left and lanelet.adj_left_same_direction:
            left_lane_available = True
            left_lane_lanelet_id = lanelet.adj_left

        right_lane_lanelet_id = -1
        right_lane_available = False
        if lanelet.adj_right and lanelet.adj_right_same_direction:
            right_lane_available = True
            right_lane_lanelet_id = lanelet.adj_right

        neighbour_lanes = NeighbourLanes(
            left_lane_available,
            left_lane_lanelet_id,
            right_lane_available,
            right_lane_lanelet_id,
        )

        return neighbour_lanes

    # -- Surrounding vehicles --

    def find_surrounding_vehicles(
        self, ego_x: float, ego_y: float, dynamic_objects: dict
    ) -> SurroundingVehicles:
        ego_lanelet_id = self.find_lanelet_id_by_position(ego_x, ego_y)
        ego_ref_s, _ = self.from_cart_to_ref_frenet(ego_x, ego_y)

        # Map all dynamic objects to lanelet ids
        lanelet_id_to_do_id = {}
        do_id_to_lanelet_id = {}
        for do_id, do in dynamic_objects.items():
            do_lanelet_id = self.find_lanelet_id_by_position(do["x"], do["y"])
            # Add to do dict
            do["lanelet_id"] = do_lanelet_id
            # Add to mapping dicts
            if do_lanelet_id not in lanelet_id_to_do_id:
                lanelet_id_to_do_id[do_lanelet_id] = []
            lanelet_id_to_do_id[do_lanelet_id].append(do_id)

            do_id_to_lanelet_id[do_id] = do_lanelet_id

        # -- Identify surrounding objects ---

        # Identify lead object
        lead_vehicle = self._find_lead_vehicle(
            ego_ref_s, ego_lanelet_id, dynamic_objects, lanelet_id_to_do_id
        )

        # Left front etc..
        left_lead_vehicle, left_rear_vehicle = self._find_adjacent_vehicles(
            "left", ego_ref_s, ego_lanelet_id, dynamic_objects, lanelet_id_to_do_id
        )
        right_lead_vehicle, right_rear_vehicle = self._find_adjacent_vehicles(
            "right", ego_ref_s, ego_lanelet_id, dynamic_objects, lanelet_id_to_do_id
        )

        surrounding_vehicles = SurroundingVehicles(
            lead_vehicle,
            left_lead_vehicle,
            left_rear_vehicle,
            right_lead_vehicle,
            right_rear_vehicle,
        )

        return surrounding_vehicles

    def _find_lead_vehicle(
        self,
        ego_ref_s: float,
        ego_lanelet_id: int,
        dynamic_objects: dict,
        lanelet_id_to_do_id: dict,
    ) -> dict:
        # Init
        lead_vehicle_id = None
        lead_dx = None
        lead_obj = None

        if not lanelet_id_to_do_id:
            return None

        # Find objects in ego lanelet (or TODO: its successors)
        relevant_do_ids = lanelet_id_to_do_id.get(ego_lanelet_id, [])
        for do_id in relevant_do_ids:
            do = dynamic_objects[do_id]

            do_s, _ = self.from_cart_to_ref_frenet(do["x"], do["y"])

            if do_s > ego_ref_s:
                dx = do_s - ego_ref_s

                if lead_dx is None or dx < lead_dx:
                    lead_vehicle_id = do_id
                    lead_dx = dx
                    lead_obj = do

        # Find in following lanelets of ego lanelet
        if lead_vehicle_id is None:
            ego_lanelet = self.get_lanelet_by_id(ego_lanelet_id)

            ego_lanelet_successor = ego_lanelet.successor

            if ego_lanelet_successor:
                msg = "Please implement"
                raise NotImplementedError(msg)

        return lead_obj

    def _find_adjacent_vehicles(
        self,
        side: str,
        ego_ref_s: float,
        ego_lanelet_id: int,
        dynamic_objects: dict,
        lanelet_id_to_do_id: dict,
    ) -> tuple[dict, dict]:
        """
        Can only handle one set of parallel lanelets.
        Does not look at consecutive lanelets.
        """

        if side not in ("left", "right"):
            msg = "side must be left or right"
            raise ValueError(msg)

        ego_lanelet = self.get_lanelet_by_id(ego_lanelet_id)

        adj_lanelet_same_direction_map = {
            "left": ego_lanelet.adj_left_same_direction,
            "right": ego_lanelet.adj_right_same_direction,
        }

        adj_lanelet_id_map = {
            "left": ego_lanelet.adj_left,
            "right": ego_lanelet.adj_right,
        }

        adj_lead_id = None
        adj_lead_dx = None
        adj_lead_obj = None

        adj_rear_id = None
        adj_rear_dx = None
        adj_rear_obj = None

        if adj_lanelet_same_direction_map[side]:
            adj_lanelet_id = adj_lanelet_id_map[side]
            relevant_do_ids = lanelet_id_to_do_id.get(adj_lanelet_id, [])

            for do_id in relevant_do_ids:
                do = dynamic_objects[do_id]

                do_s, _ = self.from_cart_to_ref_frenet(do["x"], do["y"])

                if do_s >= ego_ref_s:
                    dx = do_s - ego_ref_s

                    if adj_lead_id is None or dx < adj_lead_dx:
                        adj_lead_id = do_id
                        adj_lead_dx = dx
                        adj_lead_obj = do
                else:
                    dx = ego_ref_s - do_s

                    if adj_rear_id is None or dx < adj_rear_dx:
                        adj_rear_id = do_id
                        adj_rear_dx = dx
                        adj_rear_obj = do

        return adj_lead_obj, adj_rear_obj

    # -- Centerlines --

    @property
    def ref_line(self) -> np.ndarray:
        return deepcopy(self._ref_line)

    def get_llt_centerline(self, llt_id: int) -> np.ndarray:
        self._check_llt_id(llt_id)
        return deepcopy(self._lanelets_by_llt_id[llt_id].center_vertices)

    # -- Reference frenet coordinate system --

    def point_array_from_cart_to_ref_frenet(
        self, pt_array_cart: np.ndarray
    ) -> np.ndarray:
        return self._pt_array_to_frenet(self._frenet_frames["ref"], pt_array_cart)

    def point_array_from_ref_frenet_to_cart(
        self, pt_array_frenet: np.ndarray
    ) -> np.ndarray:
        return self._pt_array_to_cart(self._frenet_frames["ref"], pt_array_frenet)

    def from_cart_to_ref_frenet(self, x: float, y: float) -> tuple[float, float]:
        return self._to_frenet(self._frenet_frames["ref"], x, y)

    def from_ref_frenet_to_cart(self, s: float, t: float) -> tuple[float, float]:
        return self._to_cart(self._frenet_frames["ref"], s, t)

    # -- Lanelet centerline frenet coordinate systems --

    def point_array_from_cart_to_llt_frenet(
        self, llt_id: int, pt_array_cart: np.ndarray
    ) -> np.ndarray:
        self._check_llt_id(llt_id)

        frenet_frame = self._get_llt_frenet_frame(llt_id)

        pt_array_frenet = self._pt_array_to_frenet(frenet_frame, pt_array_cart)

        return pt_array_frenet

    def point_array_from_llt_frenet_to_cart(
        self, llt_id: int, pt_array_frenet: np.ndarray
    ) -> np.ndarray:
        self._check_llt_id(llt_id)

        frenet_frame = self._get_llt_frenet_frame(llt_id)

        pt_array_cart = self._pt_array_to_cart(frenet_frame, pt_array_frenet)

        return pt_array_cart

    def from_cart_to_llt_frenet(
        self, llt_id: int, x: float, y: float
    ) -> tuple[float, float]:
        self._check_llt_id(llt_id)

        frenet_frame = self._get_llt_frenet_frame(llt_id)

        s, t = self._to_frenet(frenet_frame, x, y)

        return s, t

    def from_llt_frenet_to_cart(
        self, llt_id: int, s: float, t: float
    ) -> tuple[float, float]:
        self._check_llt_id(llt_id)

        frenet_frame = self._get_llt_frenet_frame(llt_id)

        x, y = self._to_cart(frenet_frame, s, t)

        return x, y

    # -- Heading --

    def heading_along_llt(self, llt_id: int, s_array: np.ndarray) -> np.ndarray:
        self._check_llt_id(llt_id)
        self._check_1d_array(s_array)

        frenet_cs = self._frenet_frames[llt_id]

        tangents = np.array([frenet_cs.tangent(s) for s in s_array])
        tangent_headings = np.arctan2(tangents[:, 1], tangents[:, 0])

        return tangent_headings

    # -- Internal --

    def _check_llt_id(self, llt_id: int) -> bool:
        if llt_id not in self._lanelets_by_llt_id:
            msg = f"Invalid llt_id: {llt_id}"
            raise ValueError(msg)
        return True

    def _get_llt_frenet_frame(self, llt_id: int) -> CurvilinearCoordinateSystem:
        if llt_id not in self._frenet_frames:
            frenet_frame = self._create_frenet_frame(
                self._lanelets_by_llt_id[llt_id].center_vertices
            )
            self._frenet_frames[llt_id] = frenet_frame

        return self._frenet_frames[llt_id]

    @staticmethod
    def _create_frenet_frame(
        polyline: np.ndarray, sampling_dist: float = 0.5
    ) -> CurvilinearCoordinateSystem:
        LaneletNetworkWrapper._check_2d_array(polyline)

        reference_line = resample_polyline(polyline, sampling_dist)

        frenet_frame = CurvilinearCoordinateSystem(reference_line)

        frenet_frame.compute_and_set_curvature()

        return frenet_frame

    @staticmethod
    def _pt_array_to_cart(
        frenet_cs: CurvilinearCoordinateSystem, pt_array_frenet: np.ndarray
    ) -> np.ndarray:
        LaneletNetworkWrapper._check_2d_array(pt_array_frenet)

        pt_array_cart = []
        for p in pt_array_frenet:
            x, y = LaneletNetworkWrapper._to_cart(frenet_cs, p[0], p[1])
            pt_array_cart.append([x, y])
        pt_array_cart = np.array(pt_array_cart)

        return pt_array_cart

    @staticmethod
    def _pt_array_to_frenet(
        frenet_cs: CurvilinearCoordinateSystem, pt_array_cart: np.ndarray
    ) -> np.ndarray:
        LaneletNetworkWrapper._check_2d_array(pt_array_cart)

        pt_array_frenet = []
        for p in pt_array_cart:
            x, y = LaneletNetworkWrapper._to_frenet(frenet_cs, p[0], p[1])
            pt_array_frenet.append([x, y])
        pt_array_frenet = np.array(pt_array_frenet)

        return pt_array_frenet

    @staticmethod
    def _check_1d_array(candidate: np.ndarray) -> bool:
        if not (isinstance(candidate, np.ndarray) and len(candidate.shape) == 1):
            msg = "Input must be a numpy array with shape: (X,)."
            raise Exception(msg)
        return True

    @staticmethod
    def _check_2d_array(candidate: np.ndarray) -> bool:
        if not (
            isinstance(candidate, np.ndarray)
            and len(candidate.shape) == 2
            and candidate.shape[1] == 2
        ):
            msg = "Input must be a numpy array with shape: (X, 2)."
            raise Exception(msg)
        return True

    @staticmethod
    def _to_cart(
        frenet_cs: CurvilinearCoordinateSystem, s: float, t: float
    ) -> tuple[float, float]:
        if s == 0:
            s = 1e-3

        if s > frenet_cs.length():
            msg = f"s is too high ({s} > {frenet_cs.length()})"
            raise OutOfProjectionDomainError(msg)

        plist = frenet_cs.convert_list_of_points_to_cartesian_coords([[s, t]], 1)
        x = plist[0][0]
        y = plist[0][1]

        x = LaneletNetworkWrapper.round_to_cm(x)
        y = LaneletNetworkWrapper.round_to_cm(y)

        return x, y

    @staticmethod
    def _to_frenet(
        frenet_cs: CurvilinearCoordinateSystem, x: float, y: float
    ) -> tuple[float, float]:
        plist = frenet_cs.convert_list_of_points_to_curvilinear_coords([[x, y]], 1)
        s = plist[0][0]
        t = plist[0][1]

        s = LaneletNetworkWrapper.round_to_cm(s)
        t = LaneletNetworkWrapper.round_to_cm(t)

        return s, t

    @staticmethod
    def round_to_cm(a: float) -> float:
        """Round to 1cm precision."""
        return np.round(a, decimals=2)



import pytest
from pathlib import Path

from simple_scenario import CR_AVAILABLE, LXD_AVAILABLE

if LXD_AVAILABLE and CR_AVAILABLE:
    from simple_scenario.lxd import HighdExtractor


class TestSimpleScenarioExtractor:
    DATA_DIR = Path(__file__).parent / "assets/highD-dataset-v1.0"

    RESULT_DIR = Path(__file__).parent / "results" / "test_simple_scenario_extractor"
    RESULT_DIR.mkdir(exist_ok=True)

    @pytest.mark.skipif(
        not (CR_AVAILABLE and LXD_AVAILABLE),
        reason="commonroad and lxd extras are not installed",
    )
    def test_simple_scenario_extraction_highd(self):
        result_dir = self.RESULT_DIR / "test_simple_scenario_extraction_highd"
        result_dir.mkdir(exist_ok=True)

        extractor = HighdExtractor(self.DATA_DIR)

        simple_scenarios = extractor.extract_simple_scenarios()

        _, s, _ = next(simple_scenarios[1])

        s.get_cr_interface()

        s.render(result_dir, dpi=600)
        s.render_gif(result_dir, dpi=600)


if __name__ == "__main__":
    tester = TestSimpleScenarioExtractor()
    tester.test_simple_scenario_extraction_highd()



from pathlib import Path

from simple_scenario import Scenario
from simple_scenario.road import Road, StraightSegment, ArcSegment, ClothoidSegment


class TestOpenxExport:
    DATA_DIR = Path(__file__).parent / "assets"
    RESULT_DIR = Path(__file__).parent / "results" / "test_openx_export"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_odr_straight_road(self):
        result_dir = self.RESULT_DIR / "test_odr_straight_road"
        result_dir.mkdir(exist_ok=True)

        straight_segment = StraightSegment(1000)

        segments = [straight_segment]

        road = Road(3, 3.75, segments)

        odr = road.create_opendrive_map()

        path_to_xodr = str(result_dir / "straight.xodr")
        odr.write_xml(path_to_xodr)

    def test_odr_curved_road(self):
        result_dir = self.RESULT_DIR / "test_odr_curved_road"
        result_dir.mkdir(exist_ok=True)

        straight_segment = StraightSegment(100)
        clothoid_segment = ClothoidSegment(3)
        arc_segment = ArcSegment(1000)

        segments = [straight_segment, clothoid_segment, arc_segment]

        road = Road(3, 3.75, segments)

        odr = road.create_opendrive_map()

        path_to_xodr = str(result_dir / "curved.xodr")
        odr.write_xml(path_to_xodr)

    def test_openx_export_a_example(self):
        """
        Download result dir, place it into $ESMINI_DIR/resources and run
        esmini --osc "$ESMINI_DIR\resources\test_openx_export_a_example\a_example.xosc" --window 60 60 1024 576
        """

        result_dir = self.RESULT_DIR / "test_openx_export_a_example"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "a_example.json"

        scenario = Scenario.from_config_file(config_file)

        scenario.render(result_dir)

        scenario.save(result_dir, mode="openx")

    def test_openx_export_b_example(self):
        """
        Download result dir, place it into $ESMINI_DIR/resources and run
        esmini --osc "$ESMINI_DIR\resources\test_openx_export_b_example\b_example.xosc" --window 60 60 1024 576
        """

        result_dir = self.RESULT_DIR / "test_openx_export_b_example"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "b_example.json"

        scenario = Scenario.from_config_file(config_file)

        scenario.render(result_dir)

        scenario.save(result_dir, mode="openx")

    def test_openx_export_cutout_example(self):
        """
        Download result dir, place it into $ESMINI_DIR/resources and run
        esmini --osc "$ESMINI_DIR\resources\test_openx_export_b_example\b_example.xosc" --window 60 60 1024 576
        """

        result_dir = self.RESULT_DIR / "test_openx_export_cutout_example"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "cutout_example.json"

        scenario = Scenario.from_config_file(config_file)

        scenario.render(result_dir)

        scenario.save(result_dir, mode="openx")


if __name__ == "__main__":
    tester = TestOpenxExport()
    tester.test_openx_export_cutout_example()



import matplotlib.pyplot as plt
import numpy as np

from matplotlib.patches import Rectangle

from .road import Road
from .rendering import Renderable
from .vehicle import Vehicle


class EgoConfiguration(Renderable):
    def __init__(
        self,
        lanelet_id: int,
        s0: float,
        t0: float,
        v0: float,
        vehicle_type_name: str = "medium",
    ) -> None:
        """
        lanelet_id: ID of initial lanelet the ego vehicle is starting on
        s0: Initial longitudinal position relative to the given lanelet in m
        t0: Initial lateral offset to the center line of the given lanelet in m
        v0: Initial speed in m/s
        vehicle_type: Type string of the vehicle.
        """

        self._config = {
            "lanelet_id": lanelet_id,
            "s0": s0,
            "t0": t0,
            "v0": v0,
            "vehicle_type_name": vehicle_type_name,
        }

        self._lanelet_id = lanelet_id
        self._s0 = s0
        self._t0 = t0
        self._v0 = v0
        self._vehicle_type_name = vehicle_type_name

        # Create dummy vehicle to get vehicle length and width (for thw calculation)
        self._vehicle = Vehicle(
            -1,
            self._lanelet_id,
            self._s0,
            self._t0,
            self._v0,
            vehicle_type_name=self._vehicle_type_name,
        )

        # Compiled values
        self._compiled = False
        self._x0 = None
        self._y0 = None
        self._heading0 = None

    @property
    def config(self) -> dict:
        return self._config

    @property
    def lanelet_id(self) -> int:
        return self._lanelet_id

    @property
    def s0(self) -> float:
        return self._s0

    @property
    def t0(self) -> float:
        return self._t0

    @property
    def v0(self) -> float:
        return self._v0

    @property
    def length(self) -> float:
        return self._vehicle.length

    @property
    def width(self) -> float:
        return self._vehicle.width

    @property
    def v_lon_min(self) -> float:
        return 0

    @property
    def v_lon_max(self) -> float:
        return self._vehicle.vehicle_parameters.longitudinal.v_max

    @property
    def a_lon_min(self) -> float:
        return -self._vehicle.vehicle_parameters.longitudinal.a_max

    @property
    def a_lon_max(self) -> float:
        return self._vehicle.vehicle_parameters.longitudinal.a_max

    @property
    def x0(self) -> float:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._x0

    @property
    def y0(self) -> float:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._y0

    @property
    def heading0(self) -> float:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._heading0

    @property
    def compiled(self) -> bool:
        return self._compiled

    def compile(self, road: Road) -> None:
        lanelet = road.lanelet_map.laneletLayer[self._lanelet_id]

        x0, y0 = road.from_frenet_to_cart(lanelet.centerline, self._s0, self._t0)

        # Assumption: heading is along the lanelet
        # Moved after 1s
        s1 = self._s0 + self._v0
        t1 = self._t0

        x1, y1 = road.from_frenet_to_cart(lanelet.centerline, s1, t1)

        heading0 = np.arctan2(y1 - y0, x1 - x0)

        self._x0 = x0
        self._y0 = y0
        self._heading0 = heading0

        self._compiled = True

    def _plot_in_ax(self, ax: plt.Axes, *args, **kwargs) -> None:  # noqa: ARG002
        # PLOT
        def get_rotated_bbox_pts(
            center_x: float,
            center_y: float,
            heading: float,
            length: float,
            width: float,
        ) -> np.ndarray:
            T = np.array(  # noqa: N806
                [
                    [np.cos(heading), -np.sin(heading), 0, center_x],
                    [np.sin(heading), np.cos(heading), 0, center_y],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1],
                ]
            )

            xmin = -length / 2
            xmax = length / 2
            ymin = -width / 2
            ymax = width / 2

            pt0 = np.dot(T, np.array([xmin, ymin, 0, 1]))[:2]
            pt1 = np.dot(T, np.array([xmax, ymin, 0, 1]))[:2]
            pt2 = np.dot(T, np.array([xmax, ymax, 0, 1]))[:2]
            pt3 = np.dot(T, np.array([xmin, ymax, 0, 1]))[:2]

            pts = np.array([pt0, pt1, pt2, pt3])
            return pts

        pts = get_rotated_bbox_pts(
            self._x0, self._y0, self._heading0, self.length, self.width
        )

        pts_closed = np.vstack((pts, pts[0]))
        ax.plot(pts_closed[:, 0], pts_closed[:, 1], "b-", zorder=20, label="Ego start")

        rect_patch = Rectangle(
            (self._x0 - self.length / 2, self._y0 - self.width / 2),
            self.length,
            self.width,
            angle=self._heading0,
            rotation_point="center",
            alpha=0.25,
            zorder=20,
            facecolor="b",
            edgecolor="b",
            label="Ego start",
        )
        ax.add_patch(rect_patch)

    def _format_ax(self, ax: plt.Axes, *args, **kwargs) -> None:  # noqa: ARG002
        ax.grid()
        ax.legend()

        ax.set(
            title="Ego configuration",
            xlabel="X position in m",
            ylabel="Y position in m",
        )

        ax.set_aspect("equal")



from __future__ import annotations

import numpy as np

from .road_segment import RoadSegment


class PolylineSegment(RoadSegment):
    def __init__(
        self,
        ref_line: list[float, float],
        offset_lines: dict[float, list[float, float]],
    ) -> None:
        """
        ref_line: Polyline defining the ref_line of the road segment
        offset_lines: List of polylines defining the offset lines of the road segment
        """

        self._config = {"ref_line": ref_line, "offset_lines": offset_lines}

        super().__init__()

        self._ref_line = np.array(ref_line).T
        self._offset_lines = {
            offset: np.array(offset_line).T
            for offset, offset_line in offset_lines.items()
        }

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @property
    def config(self) -> dict:
        return self._config



import numpy as np

from abc import abstractmethod


class RoadSegment:
    def __init__(self, sampling_distance: float = 0.5) -> None:
        """
        sampling_distance: Sampling distance along s in m
        """

        self._sampling_distance = sampling_distance

        self._ref_line = None
        self._ref_line_curvature = None
        self._ref_line_heading = None
        self._offset_lines = {}

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @property
    def sampling_distance(self) -> float:
        return self._sampling_distance

    @property
    def ref_line(self) -> np.ndarray:
        return self._ref_line

    @property
    def ref_line_curvature(self) -> np.ndarray:
        return self._ref_line_curvature

    @property
    def ref_line_heading(self) -> np.ndarray:
        return self._ref_line_heading

    @property
    def offset_lines(self) -> dict:
        return self._offset_lines

    @abstractmethod
    def compute_ref_line(self, **kwargs) -> np.ndarray:
        raise NotImplementedError

    def compute_offset_line(self, lateral_offset: float) -> np.ndarray:
        raise NotImplementedError



from __future__ import annotations

import numpy as np

from pyclothoids import Clothoid

from .road_segment import RoadSegment


class ClothoidSegment(RoadSegment):
    def __init__(self, r: float = 3) -> None:
        """
        r: Kennstelle of clothoid. Should always be between 1 and 3. 1: easiest to drive, 3: hardest to drive
        """
        self._config = {"r": r}

        super().__init__()

        if not (1 <= r <= 3):
            msg = f"The clothoid parameter 'r' must be between 1 and 3. Now it is: {r}"
            raise ValueError(msg)

        self._r = r

        self._ref_clothoid = None
        self._offset_clothoids = {}

        # State
        self._clothoid_start_curvature = None
        self._clothoid_end_curvature = None
        self._ref_line = None
        self._ref_line_length = None

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} r={self._r}>"

    @property
    def config(self) -> dict:
        return self._config

    @property
    def r(self) -> float:
        return self._r

    @property
    def ref_clothoid(self) -> Clothoid:
        return self._ref_clothoid

    @property
    def offset_clothoids(self) -> list[Clothoid]:
        return self._offset_clothoids

    @property
    def curvature_start(self) -> float:
        if self._clothoid_start_curvature is None:
            msg = (
                "Need to call compute_ref_line() first by adding the segment to a road."
            )
            raise RuntimeError(msg)

        return self._clothoid_start_curvature

    @property
    def curvature_end(self) -> float:
        if self._clothoid_end_curvature is None:
            msg = (
                "Need to call compute_ref_line() first by adding the segment to a road."
            )
            raise RuntimeError(msg)

        return self._clothoid_end_curvature

    @property
    def length(self) -> float:
        if self._ref_line_length is None:
            msg = (
                "Need to call compute_ref_line() first by adding the segment to a road."
            )
            raise RuntimeError(msg)

        return self._ref_line_length

    def compute_ref_line(
        self, start_x: float, start_y: float, start_heading: float, end_radius: float
    ) -> np.ndarray:
        clothoid_param_A = end_radius / self.r  # noqa: N806
        clothoid_length_L = (clothoid_param_A**2) / end_radius  # noqa: N806

        self._clothoid_start_curvature = 0.0
        self._clothoid_end_curvature = 1 / end_radius

        curvature_rate = (
            self._clothoid_end_curvature - self._clothoid_start_curvature
        ) / clothoid_length_L

        clothoid = Clothoid.StandardParams(
            start_x,
            start_y,
            start_heading,
            self._clothoid_start_curvature,
            curvature_rate,
            clothoid_length_L,
        )

        clothoid_x = []
        clothoid_y = []

        for s in np.arange(
            0, clothoid_length_L + self._sampling_distance, self._sampling_distance
        ):
            clothoid_x.append(clothoid.X(s))
            clothoid_y.append(clothoid.Y(s))

        ref_line = np.array([clothoid_x, clothoid_y]).transpose()

        self._ref_clothoid = clothoid
        self._ref_line = ref_line

        clothoid_curvature = [0.0]
        for _ in np.arange(
            self._sampling_distance,
            clothoid.length + self._sampling_distance,
            self._sampling_distance,
        ):
            clothoid_curvature.append(
                clothoid_curvature[-1] + curvature_rate * self._sampling_distance
            )
        self._ref_line_curvature = clothoid_curvature

        heading = [
            clothoid.Theta(s)
            for s in np.arange(
                0, clothoid.length + self._sampling_distance, self._sampling_distance
            )
        ]

        self._ref_line_heading = heading

        self._ref_line_length = float(
            np.sum(
                np.sqrt(
                    np.diff(self._ref_line[:, 0]) ** 2
                    + np.diff(self._ref_line[:, 1]) ** 2
                )
            )
        )

        return ref_line

    def compute_offset_line(self, lateral_offset: float) -> np.ndarray:
        clothoid_start_x = self._ref_line[0, 0] + lateral_offset * np.cos(
            self._ref_clothoid.ThetaStart - np.pi / 2
        )
        clothoid_start_y = self._ref_line[0, 1] + lateral_offset * np.sin(
            self._ref_clothoid.ThetaStart - np.pi / 2
        )

        clothoid_end_x = self._ref_line[-1, 0] + lateral_offset * np.cos(
            self._ref_clothoid.ThetaEnd - np.pi / 2
        )
        clothoid_end_y = self._ref_line[-1, 1] + lateral_offset * np.sin(
            self._ref_clothoid.ThetaEnd - np.pi / 2
        )

        clothoid = Clothoid.G1Hermite(
            clothoid_start_x,
            clothoid_start_y,
            self._ref_clothoid.ThetaStart,
            clothoid_end_x,
            clothoid_end_y,
            self._ref_clothoid.ThetaEnd,
        )

        # Make sure there are exactly the same amount of points than in ref_line
        n_ref_clothoids_points = self._ref_line.shape[0]
        offset_line = np.array(clothoid.SampleXY(n_ref_clothoids_points)).transpose()

        self._offset_clothoids[lateral_offset] = clothoid
        self._offset_lines[lateral_offset] = offset_line

        return offset_line



from __future__ import annotations

import numpy as np

from loguru import logger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Generator
    from pathlib import Path

from .. import CR_AVAILABLE, LXD_AVAILABLE

if CR_AVAILABLE:
    from ..lanelet_network_wrapper import LaneletNetworkWrapper
    from ..commonroad_interface import LaneletNetworkCreator

if LXD_AVAILABLE:
    from lxd_io import Dataset

from .. import EgoConfiguration, Vehicle, Scenario
from ..road import Road


class HighdExtractor:
    def __init__(self, dataset_dir: str | Path) -> None:
        if not (CR_AVAILABLE and LXD_AVAILABLE):
            msg = "Please install the commonroad and lxd extras to use this feature. `pip install simple_scenario[commonroad, lxd]`"
            raise ModuleNotFoundError(msg)

        self._dataset = Dataset(dataset_dir)

        self._lower_roads_per_recording = {}
        self._lower_road_wrappers_per_recording = {}
        self._upper_roads_per_recording = {}
        self._upper_road_wrappers_per_recording = {}

        self._default_speed_limit = 130
        self._start_line_s = 500 + 50  # 500 m (because the road starts at -500 m)

    def extract_simple_scenarios(self) -> dict[int, Generator[Scenario]]:
        all_scenarios = {}

        for recording_id in self._dataset.recording_ids:
            all_scenarios[recording_id] = self.extract_recording_scenarios(recording_id)

        return all_scenarios

    def get_specific_scenario(
        self, recording_id: int, track_id: int
    ) -> tuple[str, Scenario | None, str | None]:
        return self._extract_track_scenario(recording_id, track_id)

    def extract_recording_scenarios(
        self,
        recording_id: int,
        first_track_idx_to_process: int = -1,
        last_track_idx_to_process: int = -1,
    ) -> Generator[tuple[str, Scenario | None, str | None]]:
        recording = self._dataset.get_recording(recording_id)

        for track_idx, track_id in enumerate(recording.track_ids):
            # Skip track_idcs that should not be processed here.
            if (
                first_track_idx_to_process != -1
                and last_track_idx_to_process != -1
                and not (
                    first_track_idx_to_process <= track_idx <= last_track_idx_to_process
                )
            ):
                continue

            try:
                scenario_id, scenario, extraction_error_msg = (
                    self._extract_track_scenario(recording_id, track_id)
                )
            except Exception as e:  # noqa: BLE001
                logger.error(
                    f"Recording {recording.id}: Skip track with idx '{track_idx}' id '{track_id}', because of exception during extraction: {e}"
                )
                scenario_id = f"highd_r{recording_id:02d}_t{track_id:05d}_error"
                scenario = None
                extraction_error_msg = "exception_during_extraction"

            yield scenario_id, scenario, extraction_error_msg

    def _extract_track_scenario(  # noqa: PLR0912
        self, recording_id: int, track_id: int
    ) -> tuple[str, Scenario | None, str | None]:
        recording = self._dataset.get_recording(recording_id)
        recording_frames_array = np.array(
            [*recording.frames, recording.frames[-1] + 1]
        )  # append one frame, because in some recordings, the duration parameter seems to be off by 1 frame

        # Speed limit for ADS is 130 km/h.
        speed_limit = recording.get_meta_data("speedLimit")
        if speed_limit == -1:
            speed_limit = self._default_speed_limit
        else:
            speed_limit = round(speed_limit * 3.6)

        # Create Road objects
        if recording_id not in self._lower_roads_per_recording:
            lower_road = Road.from_highd_parameters(
                recording.get_meta_data("lowerLaneMarkings"),
                "lower",
                speed_limit=speed_limit,
            )
            upper_road = Road.from_highd_parameters(
                recording.get_meta_data("upperLaneMarkings"),
                "upper",
                speed_limit=speed_limit,
            )

            # Create CR lanelet networks and wrap them
            lower_road_wrapper = LaneletNetworkWrapper(
                LaneletNetworkCreator(lower_road).lanelet_network
            )
            upper_road_wrapper = LaneletNetworkWrapper(
                LaneletNetworkCreator(upper_road).lanelet_network
            )

            self._lower_roads_per_recording[recording_id] = lower_road
            self._lower_road_wrappers_per_recording[recording_id] = lower_road_wrapper
            self._upper_roads_per_recording[recording_id] = upper_road
            self._upper_road_wrappers_per_recording[recording_id] = upper_road_wrapper
        else:
            lower_road = self._lower_roads_per_recording[recording_id]
            lower_road_wrapper = self._lower_road_wrappers_per_recording[recording_id]
            upper_road = self._upper_roads_per_recording[recording_id]
            upper_road_wrapper = self._upper_road_wrappers_per_recording[recording_id]

        track = recording.get_track(track_id)

        # -- Road --

        # Find the relevant road (lower or upper)
        local_trajectory = track.get_local_trajectory()
        local_trajectory[:, -1] = -local_trajectory[
            :, -1
        ]  # Invert, because we want the y axis to point upwards (map is already flipped)

        first_position_x = local_trajectory[0, 0]
        first_position_y = local_trajectory[0, 1]

        ego_driving_direction = track.get_meta_data("drivingDirection")
        if ego_driving_direction == 1:
            relevant_road_name = "upper"
            relevant_road = upper_road
            relevant_wrapper = upper_road_wrapper
        elif ego_driving_direction == 2:
            relevant_road_name = "lower"
            relevant_road = lower_road
            relevant_wrapper = lower_road_wrapper
        else:
            msg = f"drivingDirection should be 1 (upper road) or 2 (lower road), but not {ego_driving_direction}"
            raise ValueError(msg)

        scenario_id = f"highd_r{recording_id:02d}_t{track_id:05d}_{relevant_road_name}"

        # Check whether it is starting before the start line and reaches the goal line
        last_position_x = local_trajectory[-1, 0]
        last_position_y = local_trajectory[-1, 1]
        first_position_s = relevant_wrapper.from_cart_to_ref_frenet(
            first_position_x, first_position_y
        )[0]
        last_position_s = relevant_wrapper.from_cart_to_ref_frenet(
            last_position_x, last_position_y
        )[0]

        # Remove trucks
        track_class = track.get_meta_data("class")
        if track_class != "Car":
            logger.info(
                f"Recording {recording_id:02d}: Skip track {track_id:04d}, because it is not a Car, but a {track_class}."
            )
            return scenario_id, None, "track_class_not_car"

        if (
            first_position_s > self._start_line_s
            or last_position_s < relevant_road.goal_position
        ):
            logger.info(
                f"Recording {recording_id:02d}: Skip track (id={track_id}), because it is too short."
            )
            return scenario_id, None, "track_too_short"

        logger.info(f"Extract simple scenario for track {track_id:05d}")

        # -- EgoConfiguration --
        # Create the ego configuration; starting point of the ego is the moment it crosses the start_line
        ego_ref_frenet = relevant_wrapper.point_array_from_cart_to_ref_frenet(
            local_trajectory
        )

        ego_start_idx = np.where(ego_ref_frenet[:, 0] > self._start_line_s)[0][0]
        ego_end_idx = np.where(ego_ref_frenet[:, 0] > relevant_road.goal_position)[0][0]

        ego_start_position_x = local_trajectory[ego_start_idx, 0]
        ego_start_position_y = local_trajectory[ego_start_idx, 1]
        ego_start_llt_id = relevant_wrapper.find_lanelet_id_by_position(
            ego_start_position_x, ego_start_position_y
        )

        ego_start_llt_s, ego_start_llt_t = relevant_wrapper.from_cart_to_llt_frenet(
            ego_start_llt_id, ego_start_position_x, ego_start_position_y
        )

        ego_v = track.get_data("xVelocity")
        if relevant_road_name == "upper":
            ego_v = -ego_v

        ego_start_v = ego_v[ego_start_idx]

        ego_configuration = EgoConfiguration(
            ego_start_llt_id, ego_start_llt_s, ego_start_llt_t, ego_start_v
        )

        # -- Vehicles --
        vehicles = []

        # Find frames that belong to ego_start_idx and ego_end_idx
        # From ego local to global frame
        initial_frame = track.get_meta_data("initialFrame")
        scenario_start_frame = initial_frame + ego_start_idx
        scenario_end_frame = initial_frame + ego_end_idx
        n_steps_reference = scenario_end_frame - scenario_start_frame + 1
        # Increase duration by 25% (distance is reduced from 400 to 300 meters, 25%)
        n_steps = int(np.ceil(n_steps_reference * 1.25))
        scenario_lifetime = np.zeros_like(recording_frames_array, dtype=bool)
        # The frames are not the indices of the frames array, because for highD, the first frame is 1 (and not 0)
        scenario_start_frame_idx = np.where(
            recording_frames_array == scenario_start_frame
        )[0][0]
        scenario_end_frame_idx = np.where(recording_frames_array == scenario_end_frame)[
            0
        ][0]
        scenario_lifetime[
            scenario_start_frame_idx : scenario_end_frame_idx + n_steps + 1
        ] = 1

        # Create Vehicle objects
        other_relevant_track_ids_ = []
        for frame in range(scenario_start_frame, scenario_end_frame + 1):
            other_relevant_track_ids_.extend(recording.get_track_ids_at_frame(frame))
        other_relevant_track_ids = sorted(set(other_relevant_track_ids_))

        cur_vehicle_id = 1000
        for other_track_id in other_relevant_track_ids:
            # Skip ego vehicle
            if other_track_id == track_id:
                continue

            other_track = recording.get_track(other_track_id)

            if other_track.get_meta_data("drivingDirection") != ego_driving_direction:
                continue

            # The other_track is maybe not present for the complete duration!
            # Append / prepend np.nan for unknown states
            other_track_initial_frame = other_track.get_meta_data("initialFrame")
            other_track_final_frame = other_track.get_meta_data("finalFrame")
            other_track_lifetime = np.zeros_like(recording_frames_array, dtype=bool)
            # The frames are not the indices of the frames array, because for highD, the first frame is 1 (and not 0)
            other_track_initial_frame_idx = np.where(
                recording_frames_array == other_track_initial_frame
            )[0][0]
            other_track_final_frame_idx = np.where(
                recording_frames_array == other_track_final_frame
            )[0][0]
            other_track_lifetime[
                other_track_initial_frame_idx : other_track_final_frame_idx + 1
            ] = 1

            other_track_length = other_track.get_meta_data("width")
            other_track_width = other_track.get_meta_data("height")
            other_track_x = other_track.get_data("x") + other_track_length / 2
            other_track_y = -(other_track.get_data("y") + other_track_width / 2)
            other_track_v = other_track.get_data("xVelocity")
            other_track_a = other_track.get_data("xAcceleration")
            other_track_heading = np.zeros_like(other_track_a)
            if relevant_road_name == "upper":
                other_track_v = -other_track_v
                other_track_a = -other_track_a
                other_track_heading += np.pi

            # Arrays for the whole recording lifetime
            other_track_x_whole_recording = np.nan * np.ones_like(
                recording_frames_array
            )
            other_track_x_whole_recording[other_track_lifetime] = other_track_x
            other_track_y_whole_recording = np.nan * np.ones_like(
                recording_frames_array
            )
            other_track_y_whole_recording[other_track_lifetime] = other_track_y
            other_track_v_whole_recording = np.nan * np.ones_like(
                recording_frames_array
            )
            other_track_v_whole_recording[other_track_lifetime] = other_track_v
            other_track_a_whole_recording = np.nan * np.ones_like(
                recording_frames_array
            )
            other_track_a_whole_recording[other_track_lifetime] = other_track_a
            other_track_heading_whole_recording = np.nan * np.ones_like(
                recording_frames_array
            )
            other_track_heading_whole_recording[other_track_lifetime] = (
                other_track_heading
            )

            # Take only the part that is in the scenario
            other_track_x_during_scenario = other_track_x_whole_recording[
                scenario_lifetime
            ]
            other_track_y_during_scenario = other_track_y_whole_recording[
                scenario_lifetime
            ]
            other_track_v_during_scenario = other_track_v_whole_recording[
                scenario_lifetime
            ]
            other_track_a_during_scenario = other_track_a_whole_recording[
                scenario_lifetime
            ]
            other_track_heading_during_scenario = other_track_heading_whole_recording[
                scenario_lifetime
            ]

            # Need at least two time steps with non-nan values
            if np.count_nonzero(~np.isnan(other_track_x_during_scenario)) < 2:
                continue

            vehicle_id = cur_vehicle_id
            # Find lanelet id for first time step the vehicle exists in the scenario
            # Find first non nan value
            other_track_first_valid_pos_x = other_track_x_during_scenario[
                ~np.isnan(other_track_x_during_scenario)
            ][0]
            other_track_first_valid_pos_y = other_track_y_during_scenario[
                ~np.isnan(other_track_y_during_scenario)
            ][0]
            other_track_initial_llt_id = relevant_road.find_lanelet_id_by_position(
                other_track_first_valid_pos_x, other_track_first_valid_pos_y
            )

            other_vehicle = Vehicle.from_data(
                vehicle_id,
                other_track_initial_llt_id,
                other_track_x_during_scenario,
                other_track_y_during_scenario,
                other_track_heading_during_scenario,
                other_track_v_during_scenario,
                other_track_a_during_scenario,
                vehicle_type_name="custom",
                length=other_track_length,
                width=other_track_width,
            )
            vehicles.append(other_vehicle)
            cur_vehicle_id += 1

        dt = 1 / recording.get_meta_data("frameRate")
        duration = n_steps * dt

        scenario = Scenario(
            scenario_id,
            relevant_road.copy(),
            ego_configuration,
            vehicles,
            from_data=True,
            duration=duration,
            dt=dt,
        )

        return scenario_id, scenario, None



from commonroad.scenario.traffic_sign import TrafficSign


class TraffiSignFixed(TrafficSign):
    def __str__(self) -> str:
        """Somehow the normal TrafficLight uses the str() in the xml file for the ref.."""
        return str(self.traffic_sign_id)



import json
import numpy as np
import pytest

from pathlib import Path

from simple_scenario import (
    Scenario,
    EgoConfiguration,
    Vehicle,
    CR_AVAILABLE,
)
from simple_scenario.rendering import create_scenario_plot_ax
from simple_scenario.road import Road, StraightSegment


class TestScenario:
    DATA_DIR = Path(__file__).parent / "assets"
    RESULT_DIR = Path(__file__).parent / "results" / "test_scenario"
    RESULT_DIR.mkdir(exist_ok=True)

    @staticmethod
    def _check_feasible(scenario: Scenario):
        if CR_AVAILABLE:
            assert scenario.is_feasible()
        else:
            with pytest.raises(ModuleNotFoundError):
                assert scenario.is_feasible()

    @staticmethod
    def _get_cr_interface(scenario: Scenario):
        if CR_AVAILABLE:
            scenario.get_cr_interface()
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario.get_cr_interface()

    @staticmethod
    def _save_to_file(scenario: Scenario, result_dir: Path):
        if scenario._initialized_from_data:  # noqa: SLF001
            with pytest.raises(
                ValueError,
                match="Cannot save to config if the scenario has been initialized from data.",
            ):
                scenario.save(result_dir)
            with pytest.raises(
                ValueError,
                match="Cannot save to openx if the scenario has been initialized from data.",
            ):
                scenario.save(result_dir, mode="openx")
        else:
            scenario.save(result_dir)
            scenario.save(result_dir, mode="openx")

        if CR_AVAILABLE:
            scenario.save(result_dir, mode="cr")
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario.save(result_dir, mode="cr")

    def test_scenario_creation(self):
        """
        Create a simple scenario in a pythonic way and save it to file (osm and json file!)
        """

        result_dir = self.RESULT_DIR / "test_scenario_creation"
        result_dir.mkdir(exist_ok=True)

        dt = 0.1
        duration = 10

        # Create road
        road = Road(3, 3.75, [StraightSegment(500, heading=0)], goal_position=450)

        # Create ego configuration
        ego_configuration = EgoConfiguration(1000, 50, 0, 27.78)

        # Create vehicles

        # A vehicle in front of the ego vehicle
        vehicle0_thw0 = 3
        vehicle0_s0 = ego_configuration.s0 + ego_configuration.v0 * vehicle0_thw0
        vehicle0 = Vehicle(
            0, ego_configuration.lanelet_id, vehicle0_s0, 0, ego_configuration.v0
        )

        # A vehicle in front of vehicle 0
        vehicle1_thw0 = 3
        vehicle1_s0 = vehicle0.s0 + vehicle0.v0 * vehicle1_thw0
        vehicle1 = Vehicle(1, ego_configuration.lanelet_id, vehicle1_s0, 0, vehicle0.v0)

        # A vehicle on another lane
        vehicle2 = Vehicle(2, 1001, ego_configuration.s0, 0, ego_configuration.v0)

        vehicles = [vehicle0, vehicle1, vehicle2]

        # Create scenario
        scenario = Scenario("test", road, ego_configuration, vehicles, duration, dt=dt)

        # Render overview
        scenario.render(result_dir, "default")
        scenario.render(result_dir, "clean", clean=True)
        # Render timestep
        scenario.render(result_dir, timestep=50, plot_name_suffix="timestep_50")
        # Render GIF
        scenario.render_gif(result_dir, dpi=300)

        # Test ax mode
        with create_scenario_plot_ax(
            result_dir, f"scenario_{scenario.id}_axmode"
        ) as ax:
            scenario.render(ax)

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_scenario_creation_big_road(self):
        """
        Create a simple scenario in a pythonic way and save it to file (osm and json file!)
        """

        result_dir = self.RESULT_DIR / "test_scenario_creation_big_road"
        result_dir.mkdir(exist_ok=True)

        dt = 0.1
        duration = 10

        # Create road
        road = Road(5, 3.75, [StraightSegment(500, heading=0)], goal_position=450)

        # Create ego configuration
        ego_configuration = EgoConfiguration(1000, 50, 0, 27.78)

        # Create vehicles

        # A vehicle in front of the ego vehicle
        vehicle0_thw0 = 3
        vehicle0_s0 = ego_configuration.s0 + ego_configuration.v0 * vehicle0_thw0
        vehicle0 = Vehicle(
            0, ego_configuration.lanelet_id, vehicle0_s0, 0, ego_configuration.v0
        )

        # A vehicle in front of vehicle 0
        vehicle1_thw0 = 3
        vehicle1_s0 = vehicle0.s0 + vehicle0.v0 * vehicle1_thw0
        vehicle1 = Vehicle(1, ego_configuration.lanelet_id, vehicle1_s0, 0, vehicle0.v0)

        # A vehicle on another lane
        vehicle2 = Vehicle(2, 1001, ego_configuration.s0, 0, ego_configuration.v0)

        vehicles = [vehicle0, vehicle1, vehicle2]

        # Create scenario
        scenario = Scenario(
            "test_big_road", road, ego_configuration, vehicles, duration, dt=dt
        )

        # Render overview
        scenario.render(result_dir, "default")
        scenario.render(result_dir, "clean", clean=True)
        # Render timestep
        scenario.render(result_dir, "timestep_50", timestep=50)
        # Render GIF
        scenario.render_gif(result_dir, dpi=300)

        # Test ax mode
        with create_scenario_plot_ax(
            result_dir, f"scenario_{scenario.id}_axmode"
        ) as ax:
            scenario.render(ax)

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_config(self):
        result_dir = self.RESULT_DIR / "test_load_from_config"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        with config_file.open("r") as f:
            config = json.load(f)

        scenario = Scenario.from_config(config)

        scenario.render(result_dir, "from_config")

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_config_file(self):
        result_dir = self.RESULT_DIR / "test_load_from_config_file"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        scenario = Scenario.from_config_file(config_file)

        scenario.render(result_dir, "from_config_file")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_cr_xml(self):
        result_dir = self.RESULT_DIR / "test_load_from_config_file"
        result_dir.mkdir(exist_ok=True)

        xml = self.DATA_DIR / "ZAM_ValidationScenario+No+PaperScenario1-0_0_T-0.xml"

        if CR_AVAILABLE:
            scenario = Scenario.from_cr_xml(xml)

            scenario.render(result_dir, "from_cr_xml")

            # Save to file
            self._save_to_file(scenario, result_dir)

            # Check feasibility
            self._check_feasible(scenario)
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario = Scenario.from_cr_xml(xml)

    def test_load_from_x_config(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_config"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        with config_file.open("r") as f:
            config = json.load(f)

        scenario = Scenario.from_x(config)

        scenario.render(result_dir, "from_x_config")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_x_config_file(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_config_file"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        scenario = Scenario.from_x(config_file)

        scenario.render(result_dir, "from_x_config_file")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_x_cr_xml(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_cr_xml"
        result_dir.mkdir(exist_ok=True)

        xml = self.DATA_DIR / "ZAM_ValidationScenario+No+PaperScenario1-0_0_T-0.xml"

        if CR_AVAILABLE:
            scenario = Scenario.from_x(xml)

            scenario.render(result_dir, "from_x_cr_xml")

            # Get cr interface
            self._get_cr_interface(scenario)

            # Save to file
            self._save_to_file(scenario, result_dir)

            # Check feasibility
            self._check_feasible(scenario)
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario = Scenario.from_x(xml)

    def test_load_from_x_object(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_object"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        scenario = Scenario.from_x(config_file)
        same_scenario = Scenario.from_x(scenario)

        same_scenario.render(result_dir, "test_load_from_x_object")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_scenario_cr_interface(self):
        """
        Create a simple scenario in a pythonic way and save it to file (osm and json file!)
        """

        dt = 0.1
        duration = 10

        # Create road
        road = Road(3, 3.75, [StraightSegment(500, heading=0)], goal_position=450)

        # Create ego configuration
        ego_configuration = EgoConfiguration(1000, 50, 0, 27.78)

        # Create vehicles

        # A vehicle in front of the ego vehicle
        vehicle0_thw0 = 3
        vehicle0_s0 = ego_configuration.s0 + ego_configuration.v0 * vehicle0_thw0
        vehicle0 = Vehicle(
            0, ego_configuration.lanelet_id, vehicle0_s0, 0, ego_configuration.v0
        )

        # A vehicle in front of vehicle 0
        vehicle1_thw0 = 3
        vehicle1_s0 = vehicle0.s0 + vehicle0.v0 * vehicle1_thw0
        vehicle1 = Vehicle(1, ego_configuration.lanelet_id, vehicle1_s0, 0, vehicle0.v0)

        # A vehicle on another lane
        vehicle2 = Vehicle(2, 1001, ego_configuration.s0, 0, ego_configuration.v0)

        vehicles = [vehicle0, vehicle1, vehicle2]

        # Create scenario
        scenario = Scenario("test", road, ego_configuration, vehicles, duration, dt=dt)

        # Get cr interface
        self._get_cr_interface(scenario)

        # Check feasibility
        self._check_feasible(scenario)

    def test_heading_calcuation_standstill(self):
        config_json = self.DATA_DIR / "scenario_standstill.json"
        scenario = Scenario.from_x(config_json)

        assert np.all(scenario.vehicles[0].heading) != 0
        self._get_cr_interface(scenario)
        self._check_feasible(scenario)


if __name__ == "__main__":
    tester = TestScenario()
    tester.test_heading_calcuation_standstill()



from __future__ import annotations

import matplotlib.pyplot as plt

from abc import abstractmethod
from pathlib import Path

from .scenario_plot_context_manager import create_scenario_plot_ax


class Renderable:
    def render(
        self,
        plot_dir_or_ax: Path | plt.Axes,
        plot_name: str | None = None,
        dpi: int = 600,
        figw: int = 9,
        figh: int = 9,
        clean: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        Either plot to file or into a given ax.
        """
        mode = self._derive_render_mode(plot_dir_or_ax, plot_name)

        if mode == "ax":
            self._plot_in_ax(plot_dir_or_ax, *args, **kwargs)

        elif mode == "plot_dir":
            with create_scenario_plot_ax(
                plot_dir_or_ax, plot_name, dpi=dpi, figw=figw, figh=figh, clean=clean
            ) as ax:
                self._plot_in_ax(ax, *args, **kwargs)
                self._format_ax(ax, *args, **kwargs)

    def _derive_render_mode(
        self, plot_dir_or_ax: Path | plt.Axes, plot_name: str
    ) -> None:
        if isinstance(plot_dir_or_ax, plt.Axes):
            return "ax"

        if isinstance(plot_dir_or_ax, (Path, str)):
            if not Path(plot_dir_or_ax).exists():
                msg = "Given 'plot_dir' does not exist. Please create it first."
                raise FileNotFoundError(msg)

            if plot_name is None:
                msg = "'plot_name' must be set when giving 'plot_dir' via 'plot_dir_or_ax'."
                raise ValueError(msg)

            return "plot_dir"

        msg = "Unsupported type of 'plot_dir_or_ax'"
        raise TypeError(msg)

    @abstractmethod
    def _plot_in_ax(self, ax: plt.Axes, *args, **kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def _format_ax(self, ax: plt.Axes, *args, **kwargs) -> None:
        raise NotImplementedError



from __future__ import annotations

import lanelet2
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from copy import deepcopy
from lanelet2.core import (
    getId,
    Lanelet,
    LaneletMap,
    LineString3d,
    Point3d,
    AttributeMap,
    BasicPoint2d,
)
from lanelet2.geometry import findNearest
from loguru import logger
from scenariogeneration import xodr
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

from .road_segment import RoadSegment
from .straight_segment import StraightSegment
from .clothoid_segment import ClothoidSegment
from .arc_segment import ArcSegment
from .polyline_segment import PolylineSegment
from ..rendering import Renderable, create_plot_ax


class Road(Renderable):
    ALLOWED_SEGMENT_SEQUENCES = (
        (StraightSegment,),
        (StraightSegment, ClothoidSegment, ArcSegment),
        (PolylineSegment,),
    )

    def __init__(
        self,
        n_lanes: int,
        lane_width: float,
        segments: list[RoadSegment],
        speed_limit: float = 120,
        goal_position: float | None = None,
        x0: float = 0,
        y0: float = 0,
    ) -> None:
        """
        n_lanes: Number of lanes in t-direction
        lane_width: Lane width of each of the lanes in m
        segments: List of segments in s-direction
        speed_limit: Speed limit on the road section in km/h
        goal_position: The s-position of the goal region in m
        x0: Start position (x) of the road in cartesian coordinates in m
        y0: Start position (y) of the road in cartesian coordinates in m
        """

        if not isinstance(segments, list):
            msg = "segments need to be a list"
            raise TypeError(msg)
        if len(segments) == 0:
            msg = "There needs to be at least one segment"
            raise TypeError(msg)
        if not all(isinstance(s, RoadSegment) for s in segments):
            msg = "Every segment needs to be a RoadSegment"
            raise TypeError(msg)

        self._config = {
            "n_lanes": n_lanes,
            "lane_width": lane_width,
            "segments": [s.config for s in segments],
            "speed_limit": speed_limit,
            "goal_position": goal_position,
            "x0": x0,
            "y0": y0,
        }

        self._n_lanes = n_lanes
        self._lane_width = lane_width
        self._segments = segments
        self._speed_limit = speed_limit
        self._goal_position = goal_position
        self._x0 = x0
        self._y0 = y0

        self._create_road_from_segments()

        self._overall_ref_line = None
        self._ref_line_length = None
        self._overall_ref_line_curvature = None
        self._overall_ref_line_heading = None
        self._overall_offset_lines = {}
        self._boundary_line = None

        if self._goal_position is None:
            self._goal_position = 0.8 * self.ref_line_length

        self._lanelet_map = self._create_lanelet_map()

    def copy(self) -> Road:
        road_copy = Road(
            self.n_lanes,
            self.lane_width,
            segments=deepcopy(self.segments),
            speed_limit=self.speed_limit,
            goal_position=self.goal_position,
            x0=self.x0,
            y0=self.y0,
        )
        return road_copy

    @property
    def config(self) -> dict:
        return self._config

    @classmethod
    def from_highd_parameters(
        cls,
        lane_markings_str: str,
        road_part: str,
        speed_limit: int,
        goal_position_from_end_of_road: float = -100,
    ) -> Road:
        """
        Create a Road object from the highD map parameters.
        """

        possible_road_parts = ("lower", "upper")

        if road_part not in possible_road_parts:
            raise ValueError

        dx = 5
        x_buffer = 500
        x_min = -x_buffer
        x_max = 450 + x_buffer
        x_values = np.arange(x_min, x_max + dx, dx).astype(float)

        lane_marking_offsets = [-float(elem) for elem in lane_markings_str.split(";")]

        all_linestrings = {}

        for offset in lane_marking_offsets:
            linestring_points_x = x_values.copy().tolist()
            linestring_points_y = [offset] * len(linestring_points_x)

            linestring = [linestring_points_x, linestring_points_y]
            all_linestrings[offset] = linestring

        # Add centerlines
        n_lanes = len(lane_marking_offsets) - 1
        for i_lanelet in range(n_lanes):
            left_line_offset = lane_marking_offsets[i_lanelet]
            right_line_offset = lane_marking_offsets[i_lanelet + 1]
            centerline_offset = round(
                left_line_offset + (right_line_offset - left_line_offset) / 2, 2
            )
            centerline_x = x_values.copy().tolist()
            centerline_y = [centerline_offset] * len(centerline_x)
            linestring = [centerline_x, centerline_y]

            all_linestrings[centerline_offset] = linestring
        all_linestrings = dict(sorted(all_linestrings.items()))

        if road_part == "lower":
            ref_line_y = max(lane_marking_offsets)  # leftmost in driving direction
            ref_line = all_linestrings[ref_line_y]
            offset_lines = {
                abs(round(offset - ref_line_y, 2)): linestring
                for offset, linestring in all_linestrings.items()
                if offset != ref_line_y
            }
            offset_lines = dict(sorted(offset_lines.items()))
            x0 = x_min
            y0 = ref_line_y
        else:
            # Invert all linestrings
            inverted_linestrings = {}
            for offset, linestring in all_linestrings.items():
                xvals = np.flip(linestring[0])
                new_linestring = [xvals.tolist(), linestring[1]]
                inverted_linestrings[offset] = new_linestring

            ref_line_y = min(lane_marking_offsets)  # leftmost in driving direction
            ref_line = inverted_linestrings[ref_line_y]
            offset_lines = {
                abs(round(ref_line_y - offset, 2)): linestring
                for offset, linestring in inverted_linestrings.items()
                if offset != ref_line_y
            }
            offset_lines = dict(sorted(offset_lines.items()))
            x0 = x_max
            y0 = ref_line_y

        polyline_segment = PolylineSegment(ref_line=ref_line, offset_lines=offset_lines)

        goal_position_s = x_max + goal_position_from_end_of_road
        lane_width = None

        road = cls(
            n_lanes=n_lanes,
            lane_width=lane_width,
            segments=[polyline_segment],
            speed_limit=speed_limit,
            goal_position=goal_position_s,
            x0=x0,
            y0=y0,
        )

        return road

    @property
    def n_lanes(self) -> int:
        return self._n_lanes

    @property
    def lane_width(self) -> int:
        return self._lane_width

    @property
    def speed_limit(self) -> int:
        return self._speed_limit

    @property
    def segments(self) -> list[RoadSegment]:
        return deepcopy(self._segments)

    @property
    def x0(self) -> float:
        return self._x0

    @property
    def y0(self) -> float:
        return self._y0

    def _create_lanelet_map(self) -> LaneletMap:
        """
        Generate Lanelet2 map from road lines
        """

        all_lines = self.offset_lines
        all_lines[0.0] = self.ref_line

        offsets = sorted(all_lines.keys(), reverse=True)

        # -- Linestrings --
        all_linestrings = {}

        n_linestrings = len(offsets)
        for i_linestring, offset in enumerate(offsets):
            line = all_lines[offset]

            linestring_points = [Point3d(getId(), pt[0], pt[1], 0.0) for pt in line]

            # Check if border
            if i_linestring == 0 or i_linestring == n_linestrings - 1:
                attributes_dict = {"type": "line_thin", "subtype": "solid"}
            elif i_linestring % 2 == 0:
                attributes_dict = {"type": "line_thin", "subtype": "dashed"}
            else:
                attributes_dict = {}

            new_linestring = LineString3d(
                getId(), linestring_points, AttributeMap(attributes_dict)
            )
            all_linestrings[offset] = new_linestring

        # -- Lanelets --

        all_lanelets = []

        n_lanelets = self._n_lanes

        lanelet_base_id = 1000

        for i_lanelet in range(n_lanelets):
            # Attributes
            attributes_dict = {
                "type": "lanelet",
                "subtype": "highway",
                "location": "nonurban",
                "region": "de",
                "one_way": "yes",
                "speed_limit": f"{self._speed_limit}",
            }

            # Linestrings
            left_line_offset = offsets[2 + 2 * i_lanelet]
            center_line_offset = offsets[1 + 2 * i_lanelet]
            right_line_offset = offsets[0 + 2 * i_lanelet]

            left_linestring = all_linestrings[left_line_offset]
            center_linestring = all_linestrings[center_line_offset]
            right_linestring = all_linestrings[right_line_offset]

            new_lanelet = Lanelet(
                lanelet_base_id + i_lanelet,
                left_linestring,
                right_linestring,
                AttributeMap(attributes_dict),
            )
            new_lanelet.centerline = center_linestring

            all_lanelets.append(new_lanelet)

        # -- Lanelet map --

        lanelet_map = lanelet2.core.createMapFromLanelets(all_lanelets)

        return lanelet_map

    @property
    def lanelet_map(self) -> LaneletMap:
        return self._lanelet_map

    @staticmethod
    def linestring2array(linestring: LineString3d) -> np.array:
        return np.array([[pt.x, pt.y] for pt in linestring])

    @staticmethod
    def array2linestring(array: np.array) -> LineString3d:
        return LineString3d(
            getId(), [Point3d(getId(), pt[0], pt[1], 0.0) for pt in array]
        )

    @staticmethod
    def from_frenet_to_cart(
        linestring: LineString3d, s: np.ndarray, t: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        def check_array(array_or_number: np.ndarray | float) -> np.ndarray:
            if isinstance(array_or_number, (float, int)):
                array_or_number = np.array([array_or_number]).astype(float)
            return array_or_number

        s = check_array(s)
        t = check_array(t)

        if s.shape != t.shape:
            raise Exception

        linestring2d = lanelet2.geometry.to2D(linestring)

        x = np.zeros_like(s)
        y = np.zeros_like(t)

        for i in range(s.shape[0]):
            arc = lanelet2.geometry.ArcCoordinates()
            arc.length = s[i]
            arc.distance = t[i]

            cart = lanelet2.geometry.fromArcCoordinates(linestring2d, arc)

            x[i] = cart.x
            y[i] = cart.y

        if x.shape[0] == 1:
            x = x[0]
            y = y[0]

        return x, y

    @staticmethod
    def from_cart_to_frenet(
        linestring: LineString3d, x: np.ndarray, y: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        def check_array(array_or_number: np.ndarray | float) -> np.ndarray:
            if isinstance(array_or_number, (float, int)):
                array_or_number = np.array([array_or_number]).astype(float)
            return array_or_number

        x = check_array(x)
        y = check_array(y)

        if x.shape != y.shape:
            raise Exception

        linestring2d = lanelet2.geometry.to2D(linestring)

        s = np.zeros_like(x)
        t = np.zeros_like(y)

        for i in range(x.shape[0]):
            arc = lanelet2.geometry.toArcCoordinates(
                linestring2d, BasicPoint2d(x[0], y[0])
            )

            s[i] = arc.length
            t[i] = arc.distance

        if s.shape[0] == 1:
            s = s[0]
            t = t[0]

        return s, t

    def find_lanelet_id_by_position(self, x: float, y: float) -> int:
        candidates = findNearest(self._lanelet_map.laneletLayer, BasicPoint2d(x, y), 1)

        if len(candidates) == 0 or candidates[0][0] != 0:
            return None

        llt_id = candidates[0][1].id

        return llt_id

    @property
    def goal_position(self) -> float:
        return self._goal_position

    @property
    def ref_line(self) -> np.ndarray:
        if self._overall_ref_line is None:
            segment_ref_lines = [seg.ref_line for seg in self._segments]
            overall_ref_line = np.concatenate(segment_ref_lines)
            self._overall_ref_line = overall_ref_line

        return deepcopy(self._overall_ref_line)

    @property
    def ref_line_curvature(self) -> np.ndarray:
        if self._overall_ref_line_curvature is None:
            segment_ref_line_curvatures = [
                seg.ref_line_curvature for seg in self._segments
            ]
            overall_curvature = np.concatenate(segment_ref_line_curvatures)
            self._overall_ref_line_curvature = overall_curvature

        return deepcopy(self._overall_ref_line_curvature)

    @property
    def ref_line_heading(self) -> np.ndarray:
        if self._overall_ref_line_heading is None:
            segment_ref_line_heading = [seg.ref_line_heading for seg in self._segments]
            overall_heading = np.concatenate(segment_ref_line_heading)
            self._overall_ref_line_heading = overall_heading

        return deepcopy(self._overall_ref_line_heading)

    @property
    def ref_line_length(self) -> float:
        if self._ref_line_length is None:
            self._ref_line_length = float(
                np.sum(
                    np.sqrt(
                        np.diff(self.ref_line[:, 0]) ** 2
                        + np.diff(self.ref_line[:, 1]) ** 2
                    )
                )
            )

        return self._ref_line_length

    @property
    def offset_lines(self) -> dict:
        if not self._overall_offset_lines:
            offset_line_list_per_offset = {}

            for seg in self._segments:
                for offset, line in seg.offset_lines.items():
                    if offset not in offset_line_list_per_offset:
                        offset_line_list_per_offset[offset] = []
                    offset_line_list_per_offset[offset].append(line)

            overall_offset_line_per_offset = {
                offset: np.concatenate(lines)
                for offset, lines in offset_line_list_per_offset.items()
            }
            self._overall_offset_lines = overall_offset_line_per_offset

        return deepcopy(self._overall_offset_lines)

    @property
    def boundary_line(self) -> np.ndarray:
        if self._boundary_line is None:
            outer_line0 = self._overall_ref_line

            all_t_offsets = sorted(self._overall_offset_lines.keys())
            highest_t = all_t_offsets[-1]
            outer_line1 = np.flipud(self._overall_offset_lines[highest_t])

            self._boundary_line = np.vstack(
                (outer_line0, outer_line1, outer_line0[0])
            )  # closed

        return self._boundary_line

    def _create_road_from_segments(self) -> None:
        segment_sequence = tuple([s.__class__ for s in self._segments])

        logger.debug("Road segment sequence: {}", segment_sequence)

        if segment_sequence not in self.ALLOWED_SEGMENT_SEQUENCES:
            msg = f"Give segment sequence ({segment_sequence}) is not in allowed segment sequence: ({self.ALLOWED_SEGMENT_SEQUENCES})"
            raise Exception(msg)

        if segment_sequence == self.ALLOWED_SEGMENT_SEQUENCES[0]:
            straight_segment = self._segments[0]

            logger.debug("Create overall ref_line")
            straight_ref_line = straight_segment.compute_ref_line(self._x0, self._y0)

            # Create offset lines
            lateral_offsets = []
            for i_lane in range(self._n_lanes):
                # Center line offset
                lateral_offsets.append((i_lane + 0.5) * self._lane_width)
                # Right boundary offset
                lateral_offsets.append((i_lane + 1) * self._lane_width)

            logger.debug("Create offset_lines at {}", lateral_offsets)

            for lateral_offset in lateral_offsets:
                straight_segment.compute_offset_line(lateral_offset)

        elif segment_sequence == self.ALLOWED_SEGMENT_SEQUENCES[1]:
            # Create segment objects
            straight_segment = self._segments[0]
            clothoid_segment = self._segments[1]
            arc_segment = self._segments[2]

            # Create ref_lines
            logger.debug("Create overall ref_line")
            straight_ref_line = straight_segment.compute_ref_line(self._x0, self._y0)
            clothoid_ref_line = clothoid_segment.compute_ref_line(
                straight_ref_line[-1, 0],
                straight_ref_line[-1, 1],
                straight_segment.heading,
                arc_segment.radius,
            )
            arc_segment.compute_ref_line(
                clothoid_ref_line[-1, 0],
                clothoid_ref_line[-1, 1],
                clothoid_segment.ref_clothoid.ThetaEnd,
            )

            # Create offset lines
            lateral_offsets = []
            for i_lane in range(self._n_lanes):
                # Center line offset
                lateral_offsets.append((i_lane + 0.5) * self._lane_width)
                # Right boundary offset
                lateral_offsets.append((i_lane + 1) * self._lane_width)

            logger.debug("Create offset_lines at {}", lateral_offsets)

            for lateral_offset in lateral_offsets:
                straight_segment.compute_offset_line(lateral_offset)
                clothoid_segment.compute_offset_line(lateral_offset)
                arc_segment.compute_offset_line(lateral_offset)

        elif segment_sequence == self.ALLOWED_SEGMENT_SEQUENCES[2]:
            # ref_line and offset_lines are already available, no need to do anything here
            pass

    def create_opendrive_map(self) -> xodr.OpenDrive:
        """
        Create an opendrive map from the road object.
        Use odrviewer.io to visualize it.
        """

        if self._segments == self.ALLOWED_SEGMENT_SEQUENCES[2]:
            raise NotImplementedError

        # Create odr object
        odr = xodr.OpenDrive("my_road")

        # Add segments
        odr_road_segments = []

        for segment in self._segments:
            if isinstance(segment, StraightSegment):
                odr_road_segment = xodr.Line(segment.length)
            elif isinstance(segment, ArcSegment):
                odr_road_segment = xodr.Arc(segment.curvature, segment.length)
            elif isinstance(segment, ClothoidSegment):
                odr_road_segment = xodr.Spiral(
                    segment.curvature_start, segment.curvature_end, segment.length
                )
            else:
                raise NotImplementedError

            odr_road_segments.append(odr_road_segment)

        # Create road
        road = xodr.create_road(
            odr_road_segments,
            1,
            left_lanes=0,
            right_lanes=self.n_lanes,
            lane_width=self.lane_width,
        )

        # Add road
        odr.add_road(road)

        # Magic
        odr.adjust_roads_and_lanes()

        return odr

    def _plot_in_ax(  # noqa: PLR0912
        self, ax: plt.Axes, use_lanelet: bool = True, verbose: bool = False
    ) -> None:
        # Start actual plotting
        if use_lanelet:
            plotted_lanelines = []

            for i, lanelet in enumerate(self._lanelet_map.laneletLayer):
                leftbound = self.linestring2array(lanelet.leftBound)
                centerline = self.linestring2array(lanelet.centerline)
                rightbound = self.linestring2array(lanelet.rightBound)

                if verbose:
                    ax.plot(
                        leftbound[:, 0],
                        leftbound[:, 1],
                        "--",
                        label=f"{lanelet.id} (left)",
                    )
                    ax.plot(
                        centerline[:, 0],
                        centerline[:, 1],
                        label=f"{lanelet.id} (center)",
                    )
                    ax.plot(
                        rightbound[:, 0],
                        rightbound[:, 1],
                        label=f"{lanelet.id} (right)",
                    )

                    ax.text(leftbound[-1, 0], leftbound[-1, 1], f"{lanelet.id} (left)")
                    ax.text(
                        centerline[-1, 0], centerline[-1, 1], f"{lanelet.id} (center)"
                    )
                    ax.text(
                        rightbound[-1, 0], rightbound[-1, 1], f"{lanelet.id} (right)"
                    )

                else:
                    left_label = None
                    right_label = None
                    if i == 0:
                        left_label = "Lane line"
                        right_label = None

                    clr = "k"

                    # Find shape (dashed/solid)
                    left_shape = "-"
                    if lanelet.leftBound.attributes["subtype"] == "dashed":
                        left_shape = "--"

                    right_shape = "-"
                    if lanelet.rightBound.attributes["subtype"] == "dashed":
                        right_shape = "--"

                    if lanelet.leftBound not in plotted_lanelines:
                        ax.plot(
                            leftbound[:, 0],
                            leftbound[:, 1],
                            left_shape,
                            color=clr,
                            label=left_label,
                            zorder=10,
                        )
                    if lanelet.rightBound not in plotted_lanelines:
                        ax.plot(
                            rightbound[:, 0],
                            rightbound[:, 1],
                            right_shape,
                            color=clr,
                            label=right_label,
                            zorder=10,
                        )

                    plotted_lanelines.append(lanelet.leftBound)
                    plotted_lanelines.append(lanelet.rightBound)

        elif verbose:
            colors = ["g", "b", "r"]

            for i, seg in enumerate(self._segments):
                ref_line = seg.ref_line

                ax.plot(
                    ref_line[:, 0],
                    ref_line[:, 1],
                    f"{colors[i]}-",
                    linewidth=2,
                    label=f"ref_line: {seg}",
                )
                ax.plot(
                    ref_line[0, 0],
                    ref_line[0, 1],
                    f"{colors[i]}x",
                    label=f"start of ref_line: {seg}",
                )
                ax.plot(
                    ref_line[-1, 0],
                    ref_line[-1, 1],
                    f"{colors[i]}o",
                    label=f"end of ref_line: {seg}",
                )

                for line in seg.offset_lines.values():
                    ax.plot(line[:, 0], line[:, 1], f"{colors[i]}-")
                    ax.plot(line[0, 0], line[0, 1], f"{colors[i]}x")
                    ax.plot(line[-1, 0], line[-1, 1], f"{colors[i]}o")

        else:
            # Plot refline
            ax.plot(
                self.ref_line[:, 0],
                self.ref_line[:, 1],
                "k-",
                zorder=11,
                label="Lane line",
            )

            # Plot all other
            for line in self.offset_lines.values():
                ax.plot(line[:, 0], line[:, 1], "k-", zorder=10)

        # Plot goal position
        s_goal = self._goal_position
        all_t_offsets = sorted(self._overall_offset_lines.keys())
        highest_t = all_t_offsets[-1]

        x_low, y_low = self.from_frenet_to_cart(
            self.array2linestring(self._overall_ref_line), s_goal, 0
        )
        x_high, y_high = self.from_frenet_to_cart(
            self.array2linestring(self._overall_ref_line), s_goal, -highest_t
        )

        ax.plot(
            [x_low, x_high],
            [y_low, y_high],
            "g-",
            zorder=10,
            linewidth=2 * mpl.rcParams["lines.linewidth"],
            label="Goal",
        )

    def _format_ax(
        self, ax: plt.Axes, use_lanelet: bool = True, verbose: bool = False
    ) -> None:
        # Formatting for plotting to file

        additional_info = []

        if use_lanelet:
            additional_info.append("lanelet2")
        if verbose:
            additional_info.append("verbose")

        title = f"Road ({', '.join(additional_info)})"

        ax.set(title=title, xlabel="X position in m", ylabel="Y position in m")
        ax.legend()
        # Set axis limits
        margin = 5
        # Find road boundaries
        road_boundary = self.boundary_line
        ax.set_xlim(
            np.min(road_boundary[:, 0]) - margin, np.max(road_boundary[:, 0]) + margin
        )
        ax.set_ylim(
            np.min(road_boundary[:, 1]) - margin, np.max(road_boundary[:, 1]) + margin
        )

        ax.set_aspect("equal")

    def render_curvature(self, plot_dir: Path, plot_name: str = "road") -> None:
        with create_plot_ax(plot_dir, f"{plot_name}_curvature") as ax:
            ax.plot(self.ref_line_curvature, label="Overall curvature")

            ax.set(
                xlabel="s in m",
                ylabel="Curvature in 1/m",
                title="Curvature of the ref_line along s",
            )
            ax.grid()
            ax.legend()

    def render_heading(self, plot_dir: Path, plot_name: str = "road") -> None:
        with create_plot_ax(plot_dir, f"{plot_name}_heading") as ax:
            ax.plot(np.rad2deg(self.ref_line_heading), label="Overall heading")

            ax.set(
                xlabel="s in m",
                ylabel="Heading in deg",
                title="Heading of the ref_line along s",
            )
            ax.grid()
            ax.legend()



import matplotlib.pyplot as plt
import numpy as np

from pathlib import Path

from simple_scenario.road import Road, StraightSegment, ArcSegment, ClothoidSegment


class TestRoad:
    RESULT_DIR = Path(__file__).parent / "results" / "test_road"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_straight_road(self):
        result_dir = self.RESULT_DIR / "test_straight_road"
        result_dir.mkdir(exist_ok=True)

        straight_segment = StraightSegment(100)

        road = Road(3, 3.75, [straight_segment], goal_position=80)

        plot_name = "short_road"

        road.render(result_dir, plot_name=f"{plot_name}_default")
        road.render(
            result_dir,
            plot_name=f"{plot_name}_no-lanelet_verbose",
            use_lanelet=False,
            verbose=True,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_no-lanelet_no-verbose",
            use_lanelet=False,
            verbose=False,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_lanelet_no-verbose",
            use_lanelet=True,
            verbose=False,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_lanelet_verbose",
            use_lanelet=True,
            verbose=True,
        )

        road.render_curvature(result_dir, plot_name=plot_name)
        road.render_heading(result_dir, plot_name=plot_name)

    def test_straight_road_at_angle(self):
        result_dir = self.RESULT_DIR / "test_straight_road_at_angle"
        result_dir.mkdir(exist_ok=True)

        straight_segment = StraightSegment(100, heading=np.pi / 4)

        road = Road(3, 3.75, [straight_segment])

        plot_name = "short_road_at_angle"

        road.render(result_dir, plot_name=f"{plot_name}_default")
        road.render(result_dir, plot_name=f"{plot_name}_default_clean", clean=True)
        road.render(
            result_dir,
            plot_name=f"{plot_name}_no-lanelet_verbose",
            use_lanelet=False,
            verbose=True,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_no-lanelet_no-verbose",
            use_lanelet=False,
            verbose=False,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_lanelet_no-verbose",
            use_lanelet=True,
            verbose=False,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_lanelet_verbose",
            use_lanelet=True,
            verbose=True,
        )

        road.render_curvature(result_dir, plot_name=plot_name)
        road.render_heading(result_dir, plot_name=plot_name)

    def test_curved_road(self):
        result_dir = self.RESULT_DIR / "test_curved_road"
        result_dir.mkdir(exist_ok=True)

        straight_segment = StraightSegment(100)
        clothoid_segment = ClothoidSegment(3)
        arc_segment = ArcSegment(1000)

        segments = [straight_segment, clothoid_segment, arc_segment]

        road = Road(2, 3.75, segments)

        plot_name = "curved_road"

        road.render(result_dir, plot_name=f"{plot_name}_default")
        road.render(result_dir, plot_name=f"{plot_name}_default_clean", clean=True)
        road.render(
            result_dir,
            plot_name=f"{plot_name}_no-lanelet_verbose",
            use_lanelet=False,
            verbose=True,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_no-lanelet_no-verbose",
            use_lanelet=False,
            verbose=False,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_lanelet_no-verbose",
            use_lanelet=True,
            verbose=False,
        )
        road.render(
            result_dir,
            plot_name=f"{plot_name}_lanelet_verbose",
            use_lanelet=True,
            verbose=True,
        )

        road.render_curvature(result_dir, plot_name=plot_name)
        road.render_heading(result_dir, plot_name=plot_name)

    def test_precise_plot(self):
        result_dir = self.RESULT_DIR / "test_precise_plot"
        result_dir.mkdir(exist_ok=True)

        straight_segment = StraightSegment(1000)

        road = Road(3, 3.75, [straight_segment], goal_position=80)

        plot_name = "long_road"

        # Render directly to file
        road.render(result_dir, plot_name=f"{plot_name}")
        road.render(result_dir, plot_name=f"{plot_name}_clean", clean=True)

        f, ax = plt.subplots()
        road.render(plot_dir_or_ax=ax)
        f.savefig(result_dir / "long_road_ax")

        road.render_curvature(result_dir, plot_name)
        road.render_heading(result_dir, plot_name)


if __name__ == "__main__":
    test_road = TestRoad()
    test_road.test_curved_road()



from __future__ import annotations

import numpy as np
import warnings

from loguru import logger
from typing import TYPE_CHECKING

from commonroad.common.file_reader import DynamicObstacleFactory
from commonroad.common.file_writer import CommonRoadFileWriter, OverwriteExistingFile
from commonroad.common.util import AngleInterval, Interval
from commonroad.planning.planning_problem import PlanningProblem, PlanningProblemSet
from commonroad.scenario.scenario import Scenario, ScenarioID, Tag
from commonroad.scenario.trajectory import State, Trajectory
from commonroad.scenario.obstacle import DynamicObstacle, ObstacleType
from commonroad.geometry.shape import Rectangle
from commonroad.planning.goal import GoalRegion
from commonroad.prediction.prediction import TrajectoryPrediction
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem
from commonroad_dc.geometry.util import resample_polyline

from .lanelet_network_creator import LaneletNetworkCreator

from ..lanelet_network_wrapper import LaneletNetworkWrapper

if TYPE_CHECKING:
    from pathlib import Path
    from ..ego_configuration import EgoConfiguration
    from ..road.road import Road
    from ..vehicle import Vehicle


class CommonroadInterface:
    """
    Interface to commonroad.
    Create scenario, planning problem set etc from concrete scenario
    """

    def __init__(
        self,
        simple_scenario_config: dict,
        ego_configuration: EgoConfiguration,
        vehicles: list[Vehicle],
        road: Road,
    ) -> None:
        logger.debug(
            f"Scenario '{simple_scenario_config['scenario_id']}': Create CR interface"
        )

        self._scenario_config = simple_scenario_config
        self._ego_configuration = ego_configuration
        self._vehicles = vehicles
        self._road = road

        # Result
        self._cr_scenario = None
        self._cr_planning_problem_set = None
        self._cr_planning_problem = None
        self._lanelet_network_wrapper = None
        self._ego_llt_id = None
        self._ego_llt_frenet_frame = None
        self._speed_limit_kph = None

        self._compile()

    @property
    def scenario(self) -> Scenario:
        return self._cr_scenario

    @property
    def planning_problem_set(self) -> PlanningProblemSet:
        return self._cr_planning_problem_set

    @property
    def planning_problem(self) -> PlanningProblem:
        if self._cr_planning_problem is None:
            # Extract single planning problem from set
            planning_problems = list(
                self._cr_planning_problem_set.planning_problem_dict.values()
            )
            if len(planning_problems) != 1:
                msg = "Need exactly one planning problem"
                raise NotImplementedError(msg)
            self._cr_planning_problem = planning_problems[0]
        return self._cr_planning_problem

    @property
    def lanelet_network_wrapper(self) -> LaneletNetworkWrapper:
        if self._lanelet_network_wrapper is None:
            self._lanelet_network_wrapper = LaneletNetworkWrapper(
                self._cr_scenario.lanelet_network
            )
        return self._lanelet_network_wrapper

    @property
    def ego_llt_id(self) -> int:
        if self._ego_llt_id is None:
            self._ego_llt_id = self.lanelet_network_wrapper.find_lanelet_id_by_position(
                *self.planning_problem.initial_state.position
            )
        return self._ego_llt_id

    @property
    def ego_llt_frenet_frame(self) -> CurvilinearCoordinateSystem:
        if self._ego_llt_frenet_frame is None:
            # Prepare curvilinear CS
            # Curvilinear CS ref line is the centerline of the lanelet the ego vehicle starts in (WHY NOT THE REFLINE?)
            # Find ego llt id
            ego_llt_id = self.lanelet_network_wrapper.find_lanelet_id_by_position(
                *self.planning_problem.initial_state.position
            )
            # Ego llt coordinate system
            self._ego_llt_frenet_frame = (
                self.lanelet_network_wrapper._get_llt_frenet_frame(ego_llt_id)  # noqa: SLF001
            )
        return self._ego_llt_frenet_frame

    @property
    def speed_limit_kph(self) -> float:
        if self._speed_limit_kph is None:
            self._speed_limit_kph = self.lanelet_network_wrapper.get_speed_limit_by_pos(
                *self.planning_problem.initial_state.position
            )
        return self._speed_limit_kph

    def _compile(self, configuration_id: int = 0) -> None:
        """
        More information: https://gitlab.lrz.de/tum-cps/commonroad-scenarios/-/blob/master/documentation/XML_commonRoad_2020a.pdf
        """

        # Init scenario object
        self._cr_scenario = self._create_cr_scenario_object(configuration_id)

        # Add everything to scenario
        # Layer 1: Road
        self._add_road_layer()

        # Layer 4: Moving object
        self._add_moving_objects_layer()

        # Add planning problem set
        self._cr_planning_problem_set = self._create_cr_planning_problem_set()

    def _generate_object_id(self) -> int:
        """Starts with 1."""
        return self._cr_scenario.generate_object_id()

    def _create_cr_scenario_object(self, configuration_id: int) -> Scenario:
        """Create ScenarioID object"""

        # ScenarioID
        # Test if the scenario_id is already a valid CR ID
        warnings.filterwarnings("error")
        try:
            scenario_id = ScenarioID.from_benchmark_id(
                self._scenario_config["scenario_id"], "2020a"
            )
        except UserWarning:
            map_name = f"{self._scenario_config['scenario_id']}"
            obstacle_behaviour = (
                "T"  # Single Trajectories (P=Predicted, S=Set-based Occupancy)
            )
            scenario_id = ScenarioID(
                configuration_id=configuration_id,
                map_name=map_name,
                map_id=0,
                obstacle_behavior=obstacle_behaviour,
                prediction_id=0,
            )

        warnings.filterwarnings("default")

        # Scenario
        scenario_dt = self._scenario_config["dt"]
        author = "simple_scenario"
        affiliation = "ika"
        source = "synthetical"
        tags = {Tag.HIGHWAY}  # crss.Tag.CRITICAL
        scenario = Scenario(
            scenario_dt,
            scenario_id=scenario_id,
            author=author,
            tags=tags,
            affiliation=affiliation,
            source=source,
        )

        return scenario

    def _add_road_layer(self) -> None:
        lanelet_network_creator = LaneletNetworkCreator(self._road)

        self._cr_scenario.add_objects(lanelet_network_creator.lanelet_network)

    def _add_moving_objects_layer(self) -> None:
        # Create all moving objects in one loop

        logger.debug("Create moving objects")

        if self._vehicles is None:
            logger.debug("No moving objects.")
            return

        for vehicle in self._vehicles:
            logger.debug("Create moving object with id: {}", vehicle.id)

            object_id = self._generate_object_id()

            moving_object_shape = Rectangle(width=vehicle.width, length=vehicle.length)

            initial_state = None
            moving_object_states = []

            n_timesteps = vehicle.x.shape[0]

            # Fill states
            for timestep in range(n_timesteps):
                if np.isnan(vehicle.x[timestep]):
                    continue

                state = State(
                    position=np.array([vehicle.x[timestep], vehicle.y[timestep]]),
                    velocity=vehicle.v[timestep],
                    orientation=vehicle.heading[timestep],
                    yaw_rate=0.0,
                    slip_angle=0.0,
                    time_step=int(timestep),
                )

                if initial_state is None:
                    initial_state = state
                else:
                    moving_object_states.append(state)

            initial_trajectory_timestep = moving_object_states[0].time_step
            moving_object_trajectory = Trajectory(
                initial_trajectory_timestep, moving_object_states
            )

            # assignments, took from vehicle.py lines 467ff.
            lanelet_network = self._cr_scenario.lanelet_network
            # state list assignments
            # find_obstacle_shape_lanelets modifies lanelet_network and adds obstacle ids to identified lanelets. not written to xml
            shape_lanelet_assignment = (
                DynamicObstacleFactory.find_obstacle_shape_lanelets(
                    initial_state,
                    moving_object_states,
                    lanelet_network,
                    object_id,
                    moving_object_shape,
                )
            )
            # find_obstacle_center_lanelets does not modify lanelet_network
            center_lanelet_assignment = (
                DynamicObstacleFactory.find_obstacle_center_lanelets(
                    initial_state, moving_object_states, lanelet_network
                )
            )

            rotated_shape = moving_object_shape.rotate_translate_local(
                initial_state.position, initial_state.orientation
            )
            initial_shape_lanelet_ids = set(
                lanelet_network.find_lanelet_by_shape(rotated_shape)
            )
            initial_center_lanelet_ids = set(
                lanelet_network.find_lanelet_by_position([initial_state.position])[0]
            )

            prediction = TrajectoryPrediction(
                trajectory=moving_object_trajectory,
                shape=moving_object_shape,
                center_lanelet_assignment=shape_lanelet_assignment,
                shape_lanelet_assignment=center_lanelet_assignment,
            )

            moving_object_obstacle = DynamicObstacle(
                obstacle_id=object_id,
                obstacle_type=ObstacleType.CAR,
                obstacle_shape=moving_object_shape,
                initial_state=initial_state,
                prediction=prediction,
                initial_center_lanelet_ids=initial_center_lanelet_ids,
                initial_shape_lanelet_ids=initial_shape_lanelet_ids,
            )

            self._cr_scenario.add_objects(moving_object_obstacle)

    def _create_cr_planning_problem_set(self) -> PlanningProblemSet:
        # Read from scenario info
        duration = self._scenario_config["duration"]
        dt = self._scenario_config["dt"]

        # Create
        planning_problem_id = self._generate_object_id()

        # Ego initial state
        ego_initial_state = State(
            position=np.array([self._ego_configuration.x0, self._ego_configuration.y0]),
            velocity=self._ego_configuration.v0,
            orientation=self._ego_configuration.heading0,
            yaw_rate=0.0,
            slip_angle=0.0,
            time_step=0,
            steering_angle=0.0,
            steering_angle_speed=0.0,
        )

        # Goal regions
        goal_states = []

        max_time_step = int(np.floor(duration / dt))

        # Make end of road goal region
        goal_area_length = 0.1
        goal_area_width = max(self._road.offset_lines.keys())

        # In ref_line frenet coordinates
        goal_area_pos_s = self._road.goal_position
        goal_area_pos_t = -goal_area_width / 2

        # To cart
        ref_line_frenet_cs = CurvilinearCoordinateSystem(
            resample_polyline(self._road.ref_line, 0.5)
        )
        goal_area_center_cart = np.array(
            ref_line_frenet_cs.convert_list_of_points_to_cartesian_coords(
                [[goal_area_pos_s, goal_area_pos_t]], 1
            )[0]
        )

        goal_road_tangent = ref_line_frenet_cs.tangent(goal_area_pos_s)
        goal_area_theta = np.arctan2(goal_road_tangent[1], goal_road_tangent[0])

        goal_shape = Rectangle(
            length=goal_area_length,
            width=goal_area_width,
            orientation=goal_area_theta,
            center=goal_area_center_cart,
        )

        road_end_goal_state = State(
            position=goal_shape,
            time_step=Interval(0, max_time_step),
            orientation=AngleInterval(
                goal_area_theta - np.pi / 8, goal_area_theta + np.pi / 8
            ),
        )
        goal_states.append(road_end_goal_state)

        goal_region = GoalRegion(goal_states)

        # To planning problem and pp set
        planning_problem = PlanningProblem(
            planning_problem_id, ego_initial_state, goal_region
        )
        planning_problem_set = PlanningProblemSet([planning_problem])

        self._cr_planning_problem_set = planning_problem_set

        return self._cr_planning_problem_set

    def to_xml(self, xml_save_dir: Path) -> Path:
        xml_save_path = xml_save_dir / f"{self._cr_scenario.scenario_id}.xml"

        logger.info("Write scenario to cr xml: {}", xml_save_path)
        file_writer = CommonRoadFileWriter(
            self._cr_scenario, self._cr_planning_problem_set
        )
        file_writer.write_to_file(
            str(xml_save_path), OverwriteExistingFile.SKIP, check_validity=True
        )

        return xml_save_path

    @property
    def scenario_id_str(self) -> str:
        if self._cr_scenario is None:
            return None

        return str(self._cr_scenario.scenario_id)



from loguru import logger
import sys

logger.configure(handlers=[{"sink": sys.stdout, "level": "WARNING"}])

from ._check_extra_availability import CR_AVAILABLE, LXD_AVAILABLE
from .ego_configuration import EgoConfiguration
from .from_data_config_error import FromDataConfigError
from .scenario import Scenario
from .vehicle import Vehicle

if CR_AVAILABLE:
    from .lanelet_network_wrapper import LaneletNetworkWrapper



from __future__ import annotations

import io
import json
import matplotlib.pyplot as plt
import numpy as np
import warnings

from copy import deepcopy
from loguru import logger
from matplotlib.transforms import Bbox
from pathlib import Path
from PIL import Image
from scenariogeneration import xosc

from . import CR_AVAILABLE

if CR_AVAILABLE:
    from commonroad_dc import pycrcc
    from commonroad.common.file_reader import CommonRoadFileReader
    from commonroad.common.solution import VehicleType
    from commonroad.scenario.trajectory import Trajectory, State
    from commonroad_dc.boundary import boundary
    from commonroad_dc.collision.trajectory_queries.trajectory_queries import (
        obb_enclosure_polygons_static,
    )
    from commonroad_dc.feasibility.feasibility_checker import trajectory_feasibility
    from commonroad_dc.feasibility.vehicle_dynamics import (
        VehicleDynamics,
        VehicleParameterMapping,
    )
    from shapely.errors import ShapelyDeprecationWarning

    from .lanelet_network_wrapper import LaneletNetworkWrapper
    from .commonroad_interface import CommonroadInterface

from .from_data_config_error import FromDataConfigError
from .ego_configuration import EgoConfiguration
from .rendering import Renderable, get_rcParams
from .road import Road, StraightSegment, ClothoidSegment, ArcSegment
from .vehicle import Vehicle


class Scenario(Renderable):
    COMPILE_MODES = (
        "config",  # Such that it can be used to init again with Scenario.from_config()
        "cr",  # Save to commroad xml
        "openx",
    )

    def __init__(
        self,
        scenario_id: str,
        road: Road,
        ego_configuration: EgoConfiguration,
        vehicles: list[Vehicle] | None,
        duration: float,
        dt: float = 0.1,
        from_data: bool = False,
        check_feasibility: bool = False,
    ) -> None:
        logger.info(f"Create simple scenario '{scenario_id}'")

        if abs(round(duration / dt) - duration / dt) > 1e-6:
            msg = "'duration' must be a multiple of 'dt'."
            raise ValueError(msg)

        if vehicles is None:
            vehicles = []

        vehicle_ids = [vehicle.id for vehicle in vehicles]
        if len(set(vehicle_ids)) != len(vehicle_ids):
            msg = f"There are multiple vehicles with the same id. (Unique IDs: {set(vehicle_ids)}, all IDs: {vehicle_ids})"
            raise ValueError(msg)

        # Save parameter values in config
        self._config = {
            "scenario_id": scenario_id,
            "road": road.config,
            "ego_configuration": ego_configuration.config,
            "vehicles": [v.config for v in vehicles],
            "duration": duration,
            "dt": dt,
        }

        self._scenario_id = scenario_id
        self._road = road
        self._ego_configuration = ego_configuration
        self._vehicles = vehicles
        self._duration = duration
        self._dt = dt

        # CR interface
        self._cr_interface = None

        self._compiled = False
        self._compile()

        self._initialized_from_data = from_data

        self._is_feasible = None
        if check_feasibility:
            self._is_feasible = self.check_feasibility()

    def __str__(self) -> str:
        return f"<Scenario {self._scenario_id} ({self._duration:.2f}s, dt: {self._dt:.2f}s, {len(self._vehicles)} vehicles)>"

    def __repr__(self) -> str:
        return self.__str__()

    def copy(
        self, copied_scenario_id_suffix: str = "copy", including_vehicles: bool = True
    ) -> Scenario:
        try:
            scenario_config_empty = deepcopy(self.config)
            scenario_config_empty["scenario_id"] += f"_{copied_scenario_id_suffix}"
            if not including_vehicles:
                scenario_config_empty["vehicles"] = []
            scenario_copy = Scenario.from_config(scenario_config_empty)

        except FromDataConfigError:
            # Scenario has been created from data, cannot use config
            road_copy = self._road.copy()

            vehicles = None
            if including_vehicles:
                vehicles = deepcopy(self._vehicles)

            scenario_copy = Scenario(
                scenario_id=f"{self._scenario_id}_{copied_scenario_id_suffix}",
                road=road_copy,
                ego_configuration=deepcopy(self._ego_configuration),
                vehicles=vehicles,
                duration=self._duration,
                dt=self._dt,
                from_data=True,
            )
        return scenario_copy

    @property
    def config(self) -> dict:
        if self._initialized_from_data:
            msg = "Not possible to access config when initialized from data."
            raise FromDataConfigError(msg)
        return self._config

    @classmethod
    def from_x(
        cls, scenario_or_config_or_file: Scenario | Path | str | dict
    ) -> Scenario:
        """
        Magic function that determines input.
        """

        if isinstance(scenario_or_config_or_file, cls):
            scenario = scenario_or_config_or_file

        elif isinstance(scenario_or_config_or_file, dict):
            scenario = cls.from_config(scenario_or_config_or_file)

        elif isinstance(scenario_or_config_or_file, (Path, str)):
            scenario_or_config_or_file = Path(scenario_or_config_or_file)

            if scenario_or_config_or_file.suffix == ".json":
                scenario = cls.from_config_file(scenario_or_config_or_file)

            elif scenario_or_config_or_file.suffix == ".xml":
                scenario = cls.from_cr_xml(scenario_or_config_or_file)

        else:
            msg = "Invalid value for 'scenario_or_config_or_file'."
            raise TypeError(msg)

        return scenario

    @classmethod
    def from_config_file(cls, config_file: tuple[str, Path]) -> Scenario:
        """
        Load from a json config file.
        """

        if not isinstance(config_file, (Path, str)):
            msg = "Wrong input type for config_file. Please provide a str or Path to the json."
            raise TypeError(msg)

        config_file = Path(config_file)

        if config_file.suffix != ".json":
            msg = "Please provid a path to a json file as config_file."
            raise TypeError(msg)

        with config_file.open("r") as f:
            config = json.load(f)

        scenario = cls.from_config(config)

        return scenario

    @classmethod
    def from_config(cls, config: dict) -> Scenario:
        """
        All input paramters of all init functions of road, vehicles etc
        """

        compiled_config = deepcopy(config)

        # Road
        if len(config["road"]["segments"]) == 1:
            segments = [StraightSegment(**config["road"]["segments"][0])]
        elif len(config["road"]["segments"]) == 3:
            segments = [
                StraightSegment(**config["road"]["segments"][0]),
                ClothoidSegment(**config["road"]["segments"][1]),
                ArcSegment(**config["road"]["segments"][2]),
            ]
        else:
            raise ValueError
        compiled_config["road"]["segments"] = segments

        road = Road(**compiled_config["road"])
        compiled_config["road"] = road

        # Ego configuration
        ego_configuration = EgoConfiguration(**config["ego_configuration"])
        compiled_config["ego_configuration"] = ego_configuration

        # Vehicles
        vehicles = [Vehicle(**vehicle_config) for vehicle_config in config["vehicles"]]

        compiled_config["vehicles"] = vehicles

        # Scenario
        scenario = cls(**compiled_config)

        return scenario

    @classmethod
    def from_cr_xml(cls, cr_xml_path: str | Path) -> Scenario:  # noqa: PLR0912
        if not isinstance(cr_xml_path, (Path, str)):
            msg = "Wrong input type for cr_xml_path. Please provide a str or Path to the xml."
            raise TypeError(msg)

        cr_xml_path = Path(cr_xml_path)

        if cr_xml_path.suffix != ".xml":
            msg = f"Wrong file type. Need xml. Have: {cr_xml_path.suffix}."
            raise ValueError(msg)

        if not cr_xml_path.exists():
            msg = f"File at cr_xml_path={cr_xml_path} does not exist."
            raise FileNotFoundError(msg)

        if not CR_AVAILABLE:
            msg = "Please install the commonroad extra to use this feature. `pip install simple_scenario[commonroad]`"
            raise ModuleNotFoundError(msg)

        # Load file
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ShapelyDeprecationWarning)
            scenario, planning_problem_set = CommonRoadFileReader(
                str(cr_xml_path)
            ).open()

        # Extract planning problem
        planning_problems = list(planning_problem_set.planning_problem_dict.values())
        if len(planning_problems) != 1:
            msg = "Need exactly one planning problem"
            raise Exception(msg)
        planning_problem = planning_problems[0]

        # Road
        lanelet_network_wrapper = LaneletNetworkWrapper(scenario.lanelet_network)

        n_lanes = len(lanelet_network_wrapper.lanelet_list)
        # Check that is it a straight road
        road_lengths = [
            lanelet.distance[-1] - lanelet.distance[0]
            for lanelet in lanelet_network_wrapper.lanelet_list
        ]
        is_straight_road = all(road_lengths == road_lengths[0])
        if not is_straight_road:
            msg = "Can only handle straight roads atm."
            raise NotImplementedError(msg)
        road_length = road_lengths[0]

        lane_widths = []
        for lanelet in lanelet_network_wrapper.lanelet_list:
            diff = lanelet.left_vertices - lanelet.right_vertices
            dist = np.sqrt(diff[:, 0] ** 2 + diff[:, 1] ** 2)
            if not np.all(dist == dist[0]):
                msg = "All lanes must have the same lane width throughout the lane."
                raise NotImplementedError(msg)

            lane_widths.append(dist[0])

        if not all(lane_widths == lane_widths[0]):
            msg = "All lanes must have the same lane width"
            raise NotImplementedError(msg)
        lane_width = lane_widths[0]

        speed_limits = []
        for lanelet in lanelet_network_wrapper.lanelet_list:
            centerline = lanelet_network_wrapper.get_llt_centerline(lanelet.lanelet_id)
            speed_limit = lanelet_network_wrapper.get_speed_limit_by_pos(
                centerline[0, 0], centerline[0, 1]
            )
            speed_limits.append(speed_limit)

        if not np.all(np.isclose(speed_limits, speed_limits[0])):
            msg = "The speed limit must be the same on all lanes"
            raise NotImplementedError(msg)
        speed_limit = speed_limits[0]

        # Starting position
        p0 = lanelet_network_wrapper.lanelet_list[-1].left_vertices[0]

        # Ego configuration
        ego_initial_pos_cart = planning_problem.initial_state.position
        ego_initial_lanelet_id = lanelet_network_wrapper.find_lanelet_id_by_position(
            *ego_initial_pos_cart
        )
        ego_initial_pos_frenet = lanelet_network_wrapper.from_cart_to_llt_frenet(
            ego_initial_lanelet_id, *ego_initial_pos_cart
        )
        ego_configuration = EgoConfiguration(
            ego_initial_lanelet_id,
            *ego_initial_pos_frenet,
            planning_problem.initial_state.velocity,
        )

        if len(planning_problem.goal.state_list) != 1:
            msg = "Planning problem must contain exactly one goal region."
            raise NotImplementedError(msg)
        goal_region = planning_problem.goal.state_list[0]
        # Assumption: A slim rectangle spanning over all lanes of the road at some s
        goal_region_x = goal_region.position.vertices[:, 0].mean()
        goal_region_s, _ = lanelet_network_wrapper.from_cart_to_ref_frenet(
            goal_region_x, 0
        )

        road = Road(
            n_lanes,
            lane_width,
            [StraightSegment(road_length)],
            speed_limit,
            goal_region_s,
            x0=p0[0],
            y0=p0[1],
        )

        # Vehicles
        if len(scenario.obstacles) == 0:
            msg = "There must be at least one vehicle."
            raise NotImplementedError(msg)

        durations = []
        vehicles = []
        for obstacle in scenario.obstacles:
            # Caution: Initial state is separate from rest of trajectory
            all_states = [
                obstacle.initial_state,
                *obstacle.prediction.trajectory.state_list,
            ]
            n_states = len(all_states)
            durations.append(n_states)

            obstacle_lanelet_id = lanelet_network_wrapper.find_lanelet_id_by_position(
                *obstacle.initial_state.position
            )

            all_x = [obstacle.initial_state.position[0]]
            all_y = [obstacle.initial_state.position[1]]
            all_heading = [obstacle.initial_state.orientation]
            all_v = [obstacle.initial_state.velocity]

            for state in obstacle.prediction.trajectory.state_list:
                all_x.append(state.position[0])
                all_y.append(state.position[1])
                all_heading.append(state.orientation)
                all_v.append(state.velocity)

            all_x = np.array(all_x)
            all_y = np.array(all_y)
            all_heading = np.array(all_heading)
            all_v = np.array(all_v)

            # Calculate acceleration
            all_a = np.diff(all_v)
            all_a = np.append(all_a, all_a[-1])

            vehicle = Vehicle.from_data(
                obstacle.obstacle_id,
                obstacle_lanelet_id,
                all_x,
                all_y,
                all_heading,
                all_v,
                all_a,
            )

            vehicles.append(vehicle)

        if len(durations) > 1 and not np.all(np.array(durations) == durations[0]):
            msg = "All vehicles must have the same amount of states."
            raise NotImplementedError(msg)
        duration_steps = durations[0]

        dt = scenario.dt
        duration = duration_steps * dt

        scenario = cls(
            str(scenario.scenario_id),
            road,
            ego_configuration,
            vehicles,
            duration,
            dt,
            from_data=True,
        )

        return scenario

    def _compile(self) -> None:
        """
        Create absolute cartesian trajectories
        """

        # Compile ego confiuration
        if not self._ego_configuration.compiled:
            self._ego_configuration.compile(self._road)

        # Compile vehicles
        for vehicle in self._vehicles:
            if not vehicle.compiled:
                vehicle.compile(self._road, self._duration, self._dt)

        self._compiled = True

    @property
    def id(self) -> str:
        return self._scenario_id

    @property
    def road(self) -> Road:
        return self._road

    @property
    def ego_configuration(self) -> EgoConfiguration:
        return self._ego_configuration

    @property
    def vehicles(self) -> list[Vehicle]:
        return self._vehicles

    @property
    def duration(self) -> float:
        return self._duration

    @property
    def dt(self) -> float:
        return self._dt

    @property
    def step_start(self) -> int:
        """
        The first step of the scenario.
        """
        return 0

    @property
    def step_end(self) -> int:
        """
        The last step of the scenario. (Index, hint: If you are using )
        """
        return self.n_steps - 1

    @property
    def n_steps(self) -> int:
        """
        Total number of steps in the scenario.
        """
        return round(self._duration / self._dt)

    @property
    def steps(self) -> np.ndarray:
        """
        Array with all steps (step indices) in the scenario.
        """
        return np.arange(self.step_start, self.step_end + 1)

    @property
    def cr_interface(self) -> None:
        msg = "The cr_interface property is deprecated, because it is slow and slows down debugging. Use get_cr_interface() instead."
        raise RuntimeError(msg)

    def get_cr_interface(self) -> CommonroadInterface:
        if not CR_AVAILABLE:
            msg = "Please install the commonroad extra for this feature: pip install simple_scenario[commonroad]"
            raise ModuleNotFoundError(msg)

        if self._cr_interface is None:
            self._cr_interface = self._create_cr_interface()
        return self._cr_interface

    def _create_cr_interface(self) -> CommonroadInterface:
        commonroad_interface = CommonroadInterface(
            self._config, self._ego_configuration, self._vehicles, self._road
        )

        return commonroad_interface

    def is_feasible(
        self, check_trajectories: bool = True, check_collision: bool = True
    ) -> bool:
        if self._is_feasible is None:
            self._is_feasible = self.check_feasibility(
                check_trajectories, check_collision
            )
        return self._is_feasible

    def check_feasibility(  # noqa: PLR0912
        self, check_trajectories: bool = True, check_collision: bool = True
    ) -> bool:
        logger.debug(f"Scenario '{self._scenario_id}': Check feasibility")

        if not CR_AVAILABLE:
            msg = "Please install the commonroad extra to use this feature. `pip install simple_scenario[commonroad]`"
            raise ModuleNotFoundError(msg)

        if check_trajectories:
            allowed_error_posx = 5e-2  # 5cm
            allowed_error_posy = 5e-2  # 5cm
            allowed_error_orientation = 3e-2

            vehicle_dynamic = VehicleDynamics.KS(VehicleType.BMW_320i)
            vehicle_parameters = VehicleParameterMapping.from_vehicle_type(
                VehicleType.BMW_320i
            )
            l_wb = vehicle_parameters.a + vehicle_parameters.b

            for vehicle in self._vehicles:
                psi_dot = (
                    np.diff(vehicle.heading, append=vehicle.heading[-1]) / self._dt
                )
                safe_v = np.nan * np.ones_like(vehicle.v)
                safe_v[~np.isclose(vehicle.v, 0, atol=1e-3)] = vehicle.v[
                    ~np.isclose(vehicle.v, 0, atol=1e-3)
                ]
                steering_angle = np.arctan(np.divide(psi_dot * l_wb, safe_v))
                steering_angle[np.isnan(steering_angle)] = 0.0

                new_state_list = []
                for i, _ in enumerate(vehicle.x):
                    new_state_list.append(
                        State(
                            position=np.array((vehicle.x[i], vehicle.y[i])),
                            steering_angle=steering_angle[i],
                            velocity=vehicle.v[i],
                            orientation=vehicle.heading[i],
                            time_step=i,
                        )
                    )

                object_trajectory = Trajectory(0, new_state_list)
                feasible, _ = trajectory_feasibility(
                    object_trajectory,
                    vehicle_dynamic,
                    self._dt,
                    e=np.array(
                        [
                            allowed_error_posx,
                            allowed_error_posy,
                            allowed_error_orientation,
                        ]
                    ),
                )

                if not feasible:
                    msg = f"scenario '{self._scenario_id}': Not feasible, because of trajectory."
                    logger.error(msg)
                    return False

        if check_collision:
            # Check collision (see https://gitlab.ika.rwth-aachen.de/lva/scenario-sim-env/-/blob/main/scenario_sim_env/simulation_core.py?ref_type=heads#L217)
            cr_scenario = self.get_cr_interface().scenario

            # Road
            road_inclusion_polygon_group = boundary.create_road_polygons(
                cr_scenario,
                method="lane_polygons",
                buffer=1,
                resample=1,
                triangulate=False,
            )
            _, road_boundary_collision_object = boundary.create_road_boundary_obstacle(
                cr_scenario
            )

            # Collision objects
            timesteps = int(self._duration / self._dt)
            for timestep in range(timesteps):
                # Create collision objects
                collision_objs = []
                for do in cr_scenario.dynamic_obstacles:
                    if timestep == 0:
                        state = do.initial_state
                    elif (
                        do.prediction.initial_time_step
                        <= timestep
                        <= do.prediction.final_time_step
                    ):
                        state = do.prediction.trajectory.state_at_time_step(timestep)
                    else:
                        msg = f"scenario '{self._scenario_id}': Dynamic obstacle is not defined for the complete scenario."
                        raise RuntimeError(msg)

                    collision_obj = pycrcc.RectOBB(
                        do.obstacle_shape.length / 2,
                        do.obstacle_shape.width / 2,
                        state.orientation,
                        state.position[0],
                        state.position[1],
                    )
                    collision_objs.append(collision_obj)

                for collision_obj in collision_objs:
                    # Check road collisiion
                    if timestep == 0:
                        is_offroad = not obb_enclosure_polygons_static(
                            road_inclusion_polygon_group, collision_obj
                        )

                    is_offroad = is_offroad or road_boundary_collision_object.collide(
                        collision_obj
                    )

                    if is_offroad:
                        msg = f"scenario '{self._scenario_id}': Not feasible, because of offroad (timestep: {timestep})."
                        logger.error(msg)
                        return False

                    # Check vehicle collision

                    for other_collision_obj in collision_objs:
                        if collision_obj is other_collision_obj:
                            continue

                        is_collision = collision_obj.collide(other_collision_obj)

                        if is_collision:
                            msg = f"scenario '{self._scenario_id}': Not feasible, because of collision (timestep: {timestep})."
                            logger.error(msg)
                            return False

        return True

    def render(
        self,
        plot_dir_or_ax: Path | plt.Axes,
        plot_name_suffix: str | None = None,
        dpi: int = 600,
        figw: int = 9,
        figh: int = 9,
        clean: bool = False,
        *args,
        **kwargs,
    ) -> None:
        plot_name = f"{self._scenario_id}"
        if plot_name_suffix:
            plot_name += f"_{plot_name_suffix}"
        super().render(
            plot_dir_or_ax,
            *args,
            plot_name=plot_name,
            dpi=dpi,
            figw=figw,
            figh=figh,
            clean=clean,
            **kwargs,
        )

    def _plot_in_ax(
        self,
        ax: plt.Axes,
        timestep: int | None = None,
        *args,  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ) -> None:
        # Plot road
        self._road.render(ax)

        # Plot ego configuration
        self._ego_configuration.render(ax)

        # Plot vehicles and their trajectories
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        for i_vehicle, vehicle in enumerate(self._vehicles):
            vehicle.render(ax, timestep=timestep, color=colors[i_vehicle % len(colors)])

    def _format_ax(
        self,
        ax: plt.Axes,
        timestep: int | None = None,
        *args,  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ) -> None:
        # Set axis limits
        margin = 10
        # Find road boundaries
        road_boundary = self._road.boundary_line
        ax.set_xlim(
            np.min(road_boundary[:, 0]) - margin, np.max(road_boundary[:, 0]) + margin
        )
        ax.set_ylim(
            np.min(road_boundary[:, 1]) - margin, np.max(road_boundary[:, 1]) + margin
        )

        title = f"Scenario '{self._scenario_id}'"
        if timestep:
            title += f" ({timestep * self._dt:.1f}s)"
        ax.set_title(title)

        ax.legend(ncols=3)

        ax.set_aspect("equal")

    def render_gif(
        self,
        save_dir: Path,
        include_traces: bool = True,
        plot_name_suffix: str = "",
        dpi: int = 300,
        crop_gif: bool = True,
    ) -> Path:
        logger.debug(f"Create gif of scenario '{self._scenario_id}'")

        save_dir = Path(save_dir)

        frames = []

        for timestep in self.steps:
            with plt.rc_context(get_rcParams(dpi=dpi, hide_ticks=True)):
                f, ax = plt.subplots()

                # Plot road
                self._road.render(ax)

                # Plot ego configuration
                self._ego_configuration.render(ax)

                # Plot vehicles and their trajectories
                colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
                for i_vehicle, vehicle in enumerate(self._vehicles):
                    vehicle.render(
                        ax,
                        timestep=timestep,
                        include_trace=include_traces,
                        color=colors[i_vehicle % len(colors)],
                    )

                # Set axis limits
                margin = 10
                # Find road boundaries
                road_boundary = self._road.boundary_line
                ax.set_xlim(
                    np.min(road_boundary[:, 0]) - margin,
                    np.max(road_boundary[:, 0]) + margin,
                )
                ax.set_ylim(
                    np.min(road_boundary[:, 1]) - margin,
                    np.max(road_boundary[:, 1]) + margin,
                )

                ax.text(
                    0.01,
                    0.99,
                    f"{np.round((timestep * self._dt), 1)}s",
                    ha="left",
                    va="top",
                    transform=ax.transAxes,
                )

                ax.set_aspect("equal")

                f.set_dpi(dpi)

                f_size_inches = f.get_size_inches()
                fw = f_size_inches[0]
                fh = f_size_inches[1]
                rect = (0, 0, fw, fh)

                if crop_gif:
                    # Create rect out of tightbox (inches)
                    tightbox = f.get_tightbbox()
                    # rect : tuple (left, bottom, right, top), default: (0, 0, 1, 1)
                    rect = np.array(
                        [
                            (tightbox.x0 / fw) * fw,
                            (tightbox.y0 / fh) * fh,
                            (tightbox.x1 / fw) * fw,
                            (tightbox.y1 / fh) * fh,
                        ]
                    )
                    rect = np.round(rect, 1)
                    rect = tuple(rect)

                with io.BytesIO() as io_buf:
                    f.savefig(
                        io_buf,
                        format="raw",
                        dpi=dpi,
                        bbox_inches=Bbox.from_extents(*rect),
                    )
                    io_buf.seek(0)
                    img_data = np.frombuffer(io_buf.getvalue(), dtype=np.uint8)

                height_px = int((rect[3] - rect[1]) * dpi)
                width_px = int((rect[2] - rect[0]) * dpi)
                img_arr = np.reshape(img_data, newshape=(height_px, width_px, -1))

                frame = Image.fromarray(img_arr)
                frames.append(frame)

                plt.close(f)

        plot_name = f"{self._scenario_id}"
        if plot_name_suffix:
            plot_name += f"_{plot_name_suffix}"

        gif_filepath = save_dir / f"{plot_name}.gif"

        logger.debug(f"Start gif file creation: {gif_filepath}")
        # Create the GIF file
        with io.BytesIO() as gif_buffer:
            frames[0].save(
                gif_buffer,
                save_all=True,
                append_images=frames[1:],
                format="gif",
                loop=0,
            )
            gif_file_content = gif_buffer.getvalue()

        with gif_filepath.open("wb") as f:
            f.write(gif_file_content)

        logger.debug(f"Successfully created gif file: {gif_filepath}")

        return gif_filepath

    def save(self, result_dir: Path, mode: str = "config") -> None:
        """
        mode: "config", "cr" or "openx"
        """

        result_dir = Path(result_dir)

        if mode not in self.COMPILE_MODES:
            msg = "Compile mode not available"
            raise ValueError(msg)

        if mode == "config":
            if self._initialized_from_data:
                msg = "Cannot save to config if the scenario has been initialized from data."
                raise ValueError(msg)

            config_file = result_dir / f"{self._scenario_id}.json"

            with config_file.open("w") as f:
                json.dump(self._config, f, indent=2)

        elif mode == "cr":
            self.get_cr_interface().to_xml(result_dir)

        elif mode == "openx":
            if self._initialized_from_data:
                msg = "Cannot save to openx if the scenario has been initialized from data."
                raise ValueError(msg)

            # CAUTION: Please read the instructions in _create_openscenario() to run in esmini

            # OpenDRIVE
            odr_path = result_dir / f"{self._scenario_id}.xodr"
            odr = self._road.create_opendrive_map()
            odr.write_xml(str(odr_path))

            # OpenSCENARIO
            osc = self._create_openscenario(odr_path)
            osc.write_xml(str(result_dir / f"{self._scenario_id}.xosc"))

    def _create_openscenario(self, odr_path: str | Path) -> xosc.Scenario:
        """
        OpenScenario for esmini
        Inspired by https://gitlab.ika.rwth-aachen.de/scenario-based-validation/smartervalidation/-/blob/MA-Warmuth/smartervalidation/simulation/esmini_simulation.py?ref_type=heads#L81

        Idea: OSC file consists of just one story that includes FollowTrajectoryActions for all vehicles in the scenario.

        xosc+xodr files need to be placed together into a subfolder in esminis resource folder

        Run by `esmini --osc "D:\Programme\esmini-2.37.10\resources\test_openx_export_default\test.xosc" --window 60 60 1024 576`
        """  # noqa: W605

        # Paths (relative to xodr dir of esmini folder)
        vehicle_catalog_path = "../xosc/Catalogs/Vehicles"
        model_blue_car_path = "../models/car_blue.osgb"
        model_red_car_path = "../models/car_red.osgb"

        # Create parameters of the scenario
        scenario_parameters = xosc.ParameterDeclarations()

        # Create catalogs
        catalog = xosc.Catalog()
        catalog.add_catalog("VehicleCatalog", vehicle_catalog_path)

        # Create road network from opendrive file
        road_network = xosc.RoadNetwork(roadfile=odr_path.name)

        # Create entities
        entities = xosc.Entities()

        # Create ego vehicle object
        ego_boundingbox = xosc.BoundingBox(
            width=self._ego_configuration.width,
            length=self._ego_configuration.length,
            height=1.8,
            x_center=2.0,
            y_center=0,
            z_center=0.9,
        )
        ego_front_axle = xosc.Axle(0.523598775598, 0.8, 1.68, 2.98, 0.4)
        ego_rear_axle = xosc.Axle(0.523598775598, 0.8, 1.68, 0, 0.4)
        ego_vehicle_object = xosc.Vehicle(
            name="car_white",
            vehicle_type=xosc.VehicleCategory.car,
            boundingbox=ego_boundingbox,
            frontaxle=ego_front_axle,
            rearaxle=ego_rear_axle,
            max_speed=self._ego_configuration.v_lon_max,
            max_acceleration=self._ego_configuration.a_lon_max,
            max_deceleration=-self._ego_configuration.a_lon_min,
        )
        ego_vehicle_object.add_property_file(model_blue_car_path)
        ego_vehicle_object.add_property("model_id", "ego")

        # Add entities
        entities.add_scenario_object("ego", ego_vehicle_object)  # , controller)

        # Create other vehicle objects
        max_acceleration = 9.81
        max_deceleration = 9.81

        vehicle_objects = {}
        for vehicle in self._vehicles:
            vehicle_boundingbox = xosc.BoundingBox(
                width=vehicle.width,
                length=vehicle.length,
                height=1.5,
                x_center=1.3,
                y_center=0,
                z_center=0.8,
            )
            vehicle_front_axle = xosc.Axle(0.523598775598, 0.8, 1.68, 2.98, 0.4)
            vehicle_rear_axle = xosc.Axle(0.523598775598, 0.8, 1.68, 0, 0.4)

            vehicle_object = xosc.Vehicle(
                name="car_red",
                vehicle_type=xosc.VehicleCategory.car,
                boundingbox=vehicle_boundingbox,
                frontaxle=vehicle_front_axle,
                rearaxle=vehicle_rear_axle,
                max_speed=69,
                max_acceleration=max_acceleration,
                max_deceleration=max_deceleration,
            )
            vehicle_object.add_property_file(model_red_car_path)
            vehicle_object.add_property("model_id", str(vehicle.id))
            vehicle_objects[vehicle.id] = vehicle_object
            entities.add_scenario_object(str(vehicle.id), vehicle_object)

        # Create the init part of the storyboard
        init = xosc.Init()
        step_time = xosc.TransitionDynamics(
            xosc.DynamicsShapes.step, xosc.DynamicsDimension.time, 1
        )

        # Ego
        ego_initial_speed_action = xosc.AbsoluteSpeedAction(
            self._ego_configuration.v0, step_time
        )
        # In Simple Scenario lane ids start with 1000 from the right (driving direction)
        # In ODR, they start with -1 from the left and get smaller to the right
        ego_initial_lane_id = (
            self._ego_configuration.lanelet_id - 1000
        ) - self._road.n_lanes
        ego_start_position_action = xosc.TeleportAction(
            xosc.LanePosition(
                self._ego_configuration.s0,
                self._ego_configuration.t0,
                ego_initial_lane_id,
                1,
            )
        )

        init.add_init_action("ego", ego_initial_speed_action)
        init.add_init_action("ego", ego_start_position_action)

        # Other vehicles
        for vehicle in self._vehicles:
            vehicle_initial_speed_action = xosc.AbsoluteSpeedAction(
                vehicle.v0, step_time
            )
            vehicle_initial_lane_id = (vehicle.lanelet_id - 1000) - self._road.n_lanes
            vehicle_initial_position_action = xosc.TeleportAction(
                xosc.LanePosition(vehicle.s0, vehicle.t0, vehicle_initial_lane_id, 1)
            )

            init.add_init_action(str(vehicle.id), vehicle_initial_speed_action)
            init.add_init_action(str(vehicle.id), vehicle_initial_position_action)

        ## Init the storyboard
        stoptrigger_storyboard = xosc.ValueTrigger(
            "StoptriggerTime",
            0,
            xosc.ConditionEdge.rising,
            xosc.SimulationTimeCondition(self._duration, xosc.Rule.greaterThan),
            triggeringpoint="stop",
        )
        storyboard = xosc.StoryBoard(init, stoptrigger_storyboard)

        ## Init the story
        storyparam = xosc.ParameterDeclarations()
        story = xosc.Story(f"Act_scenario_{self._scenario_id}", storyparam)

        # Init the Act
        act = xosc.Act(f"Act_scenario_{self._scenario_id}")

        # Add FollowTrajectoryActions for all vehicles to the event
        scenario_step_times = (self.steps * self._dt).tolist()

        for vehicle in self._vehicles:
            # Init the maneuvergroup
            maneuver_group = xosc.ManeuverGroup(f"ManeuverGroup_vehicle_{vehicle.id}")

            # Init the maneuver
            maneuver = xosc.Maneuver(f"Maneuver_vehicle_{vehicle.id}")

            # Create the event
            event = xosc.Event(f"Event_vehicle_{vehicle.id}", xosc.Priority.override)

            # Add the actor to the maneuver group
            maneuver_group.add_actor(str(vehicle.id))

            # Init trajectory
            vehicle_trajectory = xosc.Trajectory(
                f"Trajectory_vehicle_{vehicle.id}", closed=False
            )

            # Create polyline
            vehicle_positions = []
            for i, _ in enumerate(vehicle.x):
                vehicle_position = xosc.WorldPosition(
                    vehicle.x[i], vehicle.y[i], h=vehicle.heading[i]
                )
                vehicle_positions.append(vehicle_position)

            vehicle_polyline = xosc.Polyline(scenario_step_times, vehicle_positions)

            # Fill trajectory with polyline
            vehicle_trajectory.add_shape(vehicle_polyline)

            # Create follow trajectory action with trajectory, reference_domain, scale and offset need to be set such that the speed is also adjusted in the simulation (here: esmini).
            vehicle_follow_trajectory_action = xosc.FollowTrajectoryAction(
                vehicle_trajectory,
                xosc.FollowingMode.position,
                reference_domain=xosc.ReferenceContext.absolute,
                scale=1,
                offset=0,
            )

            # Add the action to the event
            event.add_action(
                f"FollowTrajectoryAction_vehicle_{vehicle.id}",
                vehicle_follow_trajectory_action,
            )

            # Add the event to the maneuver
            maneuver.add_event(event)
            # Add the maneuver to the maneuver group
            maneuver_group.add_maneuver(maneuver)

            # Add the maneuver group to the act
            act.add_maneuver_group(maneuver_group)

        # Add the act to the story
        story.add_act(act)
        # Add the story to the storyboard
        storyboard.add_story(story)

        ## Put everything together to create the scenario
        osc_scenario = xosc.Scenario(
            name=self._scenario_id,
            author="simple_scenario",
            parameters=scenario_parameters,
            entities=entities,
            storyboard=storyboard,
            roadnetwork=road_network,
            catalog=catalog,
        )

        return osc_scenario



# Check extra availability
# commonroad extra
try:
    import commonroad
    import commonroad_dc
    import shapely

    CR_AVAILABLE = True
except ModuleNotFoundError:
    CR_AVAILABLE = False

# lxd extra
try:
    import lxd_io

    LXD_AVAILABLE = True
except ModuleNotFoundError:
    LXD_AVAILABLE = False



import pytest

from pathlib import Path

from simple_scenario import CR_AVAILABLE

if CR_AVAILABLE:
    from commonroad.common.file_reader import CommonRoadFileReader

    from simple_scenario import LaneletNetworkWrapper


@pytest.fixture
def dynamic_objects_dict() -> dict:
    return {
        "right_rear": {
            "x": 20.0,
            "y": -4.0,
        },
        "right_lead": {
            "x": 100.0,
            "y": -4.0,
        },
        "left_rear": {
            "x": 20.0,
            "y": -2.0,
        },
        "left_lead": {
            "x": 100.0,
            "y": -2.0,
        },
    }


class TestLaneletNetworkWrapper:
    DATA_DIR = Path(__file__).parent / "assets"
    RESULT_DIR = Path(__file__).parent / "results"

    @pytest.mark.skipif(not CR_AVAILABLE, reason="commonroad extra is not installed")
    def test_lanelet_network_wrapper(self, dynamic_objects_dict):
        xml = self.DATA_DIR / "ZAM_ValidationScenario+No+PaperScenario1-0_0_T-0.xml"

        scenario, _ = CommonRoadFileReader(str(xml)).open()

        lanelet_network_wrapper = LaneletNetworkWrapper(scenario.lanelet_network)

        assert lanelet_network_wrapper.lanelet_list

        assert lanelet_network_wrapper.from_ref_frenet_to_cart(0, 0)

        assert len(lanelet_network_wrapper.get_lanelet_dict()) == 2

        lanelet_network_wrapper.get_lanelet_by_id(1000)

        with pytest.raises(ValueError, match=r"Invalid llt_id: \d+"):
            lanelet_network_wrapper.get_lanelet_by_id(0)

        available_neighbour_lanes = (
            lanelet_network_wrapper.find_available_neighbour_lanes(1000)
        )
        assert available_neighbour_lanes.left_lane_available
        assert available_neighbour_lanes.left_lane_lanelet_id == 1001
        assert not available_neighbour_lanes.right_lane_available
        assert available_neighbour_lanes.right_lane_lanelet_id == -1

        ego_x = 50.0
        ego_y = -4.0
        surrounding_vehicles = lanelet_network_wrapper.find_surrounding_vehicles(
            ego_x, ego_y, dynamic_objects_dict
        )

        assert surrounding_vehicles.lead is not None
        assert surrounding_vehicles.left_lead is not None
        assert surrounding_vehicles.left_rear is not None
        assert surrounding_vehicles.right_lead is None
        assert surrounding_vehicles.right_rear is None

        with pytest.raises(ValueError, match="side must be left or right"):
            lanelet_network_wrapper._find_adjacent_vehicles(  # noqa: SLF001
                "hi", None, None, None, None
            )



from pathlib import Path

from simple_scenario import EgoConfiguration
from simple_scenario.rendering import create_scenario_plot_ax
from simple_scenario.road import Road, StraightSegment


class TestEgoConfiguration:
    RESULT_DIR = Path(__file__).parent / "results" / "test_ego_configuration"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_ego_configuration(self):
        result_dir = self.RESULT_DIR / "test_ego_configuration"
        result_dir.mkdir(exist_ok=True)

        road = Road(3, 3.75, [StraightSegment(200)])

        ego_configuration = EgoConfiguration(1000, 50, 0, 27.78)

        ego_configuration.compile(road)

        ego_configuration.render(result_dir, "ego_configuration")
        ego_configuration.render(result_dir, "ego_configuration_clean", clean=True)

        with create_scenario_plot_ax(result_dir, "ego_configration_on_road") as ax:
            road.render(ax)
            ego_configuration.render(ax)

        with create_scenario_plot_ax(
            result_dir, "ego_configration_on_road_clean", clean=True
        ) as ax:
            road.render(ax)
            ego_configuration.render(ax)


if __name__ == "__main__":
    tester = TestEgoConfiguration()
    tester.test_ego_configuration()
